#include "master.h"
#include "convertstring.h"
#include "reservoir.h"
#include "mpi.h"
#include "slave.h"
#include <queue>
#include <vector>
#include "boost/date_time/local_time/local_time.hpp"

Master::Master(PowerSystem* in_system, int in_rank, int in_size)
	: system(in_system), rank(in_rank), size(in_size), upper_bound(0.0), lower_bound(0.0){
}

void Master::Initialize(Input* input, Inflow* in_inflow_data) {
	should_print_bounds = true;

	inflow_data = in_inflow_data;
	inflow_data->SeedRNG(0);
	ReadData(input);
	CreateInflowSamples();
	SendInflowSamples();
	CreateForwardInflow();
	SendForwardInflow();
	hydro_modules = system->getHydroModules();
	thermal_plants = system->getThermalPlants();
	num_modules = int(hydro_modules.size());
	num_thermal = int(system->getThermalPlants().size());
	num_slaves = size - 1;
	demand = new Demand(initial_demand);
	substage_data = new Substagedata(num_substages);
	InitializeStageData();
	BuildTree();
	outgoing_messages = new double*[num_slaves];
	for (int i = 0; i < num_slaves; i++) {
		outgoing_messages[i] = NULL;
	}
	incoming_message = new double[root->num_variables + 1];
	array_of_requests = new MPI_Request[num_slaves];
	int temp = 1;
	for (int i = 0; i < num_slaves; i++) {
		array_of_requests[i] = MPI_REQUEST_NULL;
	}
	serialized_cut = new double[1 + num_modules];
	start_time = clock();
	time_spent_in_forward_cycle = 0;
	time_spent_in_backward_cycle = 0;
	time_spent_sending_backward_messages = 0;
	time_spent_receiving_backward_messages = 0;
	total_time_spent = 0;

	if (should_print_bounds) {
		bounds_log.open("../../output/bounds.txt");
	}

	if (should_print) {
		main_log.open("../../output/master.txt");
		cuts_log.open("../../output/cuts.txt");
		receive_forward_log.open("../../output/master_received_forward.txt");
		receive_backward_log.open("../../output/master_received_backward.txt");
		inflow_data->Print();
		demand->PrintDemand(num_stages);
		substage_data->Print();
	}

	if (should_print || should_print_send_log || should_print_time_data) {
		CreateProblemLog();
	}

	for (int i = 0; i < num_slaves; i++) {
		slave_queue.push_back(i+1);
	}
	for (int i = 0; i < num_slaves; i++) {
		slave_queue.push_back(i+1);
	}

	if (should_print_send_log) {
		send_log.open("../../output/master_sent.txt");
		send_time_log.open("../../output/forward_send_time.txt");
		receive_time_log.open("../../output/forward_receive_time.txt");
	}
}

void Master::CreateInflowSamples() {
	double*** new_noise_samples = inflow_data->CreateNoiseSamples();
	inflow_data->ResetNoiseSamples(new_noise_samples);
}

void Master::CreateForwardInflow() {
	double*** new_forward_inflow = inflow_data->CreateForwardInflow();
	inflow_data->ResetForwardInflow(new_forward_inflow);
}

void Master::SendInflowSamples() {
	double*** my_samples = inflow_data->get_noise_samples();
	for (int i = 0; i < max_num_iterations; i++) {
		for (int j = 0; j < num_stages; j++) {
			MPI_Bcast(my_samples[i][j], num_samples, MPI_DOUBLE, 0, MPI_COMM_WORLD);
		}
	}
}

void Master::SendForwardInflow() {
	double*** my_forward_inflow = inflow_data->get_forward_inflow();
	for (int i = 0; i < max_num_iterations; i++) {
		for (int j = 0; j < num_stages; j++) {
			MPI_Bcast(my_forward_inflow[i][j], num_branches, MPI_DOUBLE, 0, MPI_COMM_WORLD);
		}
	}
}

void Master::CreateProblemLog(){
	problem_log.open("../../output/problem.txt");
	problem_log << num_stages << " " << num_modules << " " << num_thermal << " " << num_branches << " "
		        << num_samples << " " << num_slaves << " " << num_states << " " << min_new_cuts << " " 
				<< num_substages << " " << use_relaxation << " ";
	problem_log.close();
}

void Master::ReadData(Input* input) {
	num_stages = input->num_stages;
	num_branches = input->num_branches;
	num_samples = input->num_samples;
	min_new_cuts = input->min_new_cuts;
	num_substages = input->num_substages;
	initial_demand = input->initial_demand;
	should_print = input->should_print;
	use_relaxation = input->use_relaxation;
	num_simulations = input->num_simulations;
	max_num_iterations = input->max_num_iterations;
	should_print_send_log = input->should_print_send_log;
	should_print_time_data = input->should_print_time_data;
}

void Master::InitializeStageData() {
	stage_data = new StageData*[num_stages];

	StageData* previous = NULL;
	stage_data[0] = new StageData(0, previous, demand, num_branches, size - 1);
	stage_data[0]->Initialize();
	previous = stage_data[0];
	for (int i = 1; i < num_stages; i++){
		stage_data[i] = new StageData(i, previous, demand, num_branches, size - 1);
		stage_data[i]->Initialize();
		previous = stage_data[i];
	}
}

void Master::BuildTree() {
	root = new State(0, 0, stage_data[0], NULL);
	InitializeRoot();
	State* current_state;
	for (int i = 0; i < num_branches; i++){
		current_state = root;
		for (int j = 1; j < num_stages; j++){
			current_state = new State(i, j, stage_data[j], current_state);
			InitializeState(current_state);
		}
	}
}

void Master::InitializeRoot() {
	root->branches = new State*[num_branches];
	root->initial_reservoir = new double[num_modules];
	root->demand = root->stage_data->getDemand(0);
	root->inflow = new double[num_modules];
	root->reservoir_solution = new double[num_modules];
	root->outflow = new double[num_modules];
	root->spillage = new double[num_modules];
	root->num_variables = 1 + 3*num_modules + num_thermal;
	root->num_constraints = 1 + num_modules;
	root->tag = 1000000000;
	root->solution = new double[root->num_variables];

	for (int i = 0; i < num_modules; i++){
//		root->inflow[i] = 5; //TODO
		root->initial_reservoir[i] = hydro_modules[i]->getReservoir().getInitial();
	}

	root->stage_data->set_state(root->branch, root);
}

void Master::InitializeState(State* state) {
	if (state->stage > 1){
		state->previous_state->next_state = state;
	}
	else{
		root->branches[state->branch] = state;
//		root->relative_inflow = inflow_data->get_forward_inflow(0, 0, 0);
	}
	state->demand = state->stage_data->getDemand(state->stage);
	state->inflow = new double[num_modules];
	state->reservoir_solution = new double[num_modules];
	state->outflow = new double[num_modules];
	state->spillage = new double[num_modules];
	state->num_variables = 1 + 3*num_modules + num_thermal;
	state->num_constraints = 1 + num_modules;
	state->tag = 1000000000 + 1000*state->branch + state->stage;
	state->stage_data->set_state(state->branch, state);
	state->solution = new double[state->num_variables];

	state->branches = NULL;
	state->next_state = NULL;
}

void Master::RunSimulations() {
	for (int i = 0; i < num_simulations; i++) {
		if (i > 0) {
			Reset();
		}
		RunSimulation();
	}
	StopSimulation();

	double average_iter = 0.0;
	double average_time = 0.0;
	for (int i = 0; i < results.num_iterations.size(); i++) {
		average_iter += results.num_iterations[i];
		average_time += results.time[i];
	}
	average_iter = average_iter/results.num_iterations.size();
	average_time = average_time/results.num_iterations.size();
	std::cout << "average number of iterations is " << average_iter << std::endl
		      << "average time is " << average_time << std::endl;


	PrintResults(average_iter, average_time);
}

void Master::RunSimulation() {
	boost::posix_time::ptime begin, temp, tick, tick2, end;
	boost::posix_time::time_duration diff;

//	clock_t begin, end, temp, temp2;
//	double diff;
	tolerance = 1;
	error = 2*tolerance;
	iter = 0;
	begin = boost::posix_time::microsec_clock::local_time();
	temp = begin;
//	while (error > tolerance && iter < max_num_iterations){
	while (iter < max_num_iterations){
		tick = boost::posix_time::microsec_clock::local_time();

		ForwardSolve();

		tick2 = boost::posix_time::microsec_clock::local_time();
		diff = tick2 - tick;
		time_spent_in_forward_cycle += diff.total_milliseconds();
		tick = boost::posix_time::microsec_clock::local_time();

		BackwardSolve();

		tick2 = boost::posix_time::microsec_clock::local_time();
		diff = tick2 - tick;
		time_spent_in_backward_cycle += diff.total_milliseconds();

		if (should_print) {
			for (int i = 0; i < num_stages - 1; i++) {
				main_log << stage_data[i]->getNumCuts() << ";";
			}
			main_log << std::endl;
		}

		FindBounds();

		tolerance = 1.96*sqrt(upper_bound_var);
		error = upper_bound - lower_bound;
		iter++;
		end = boost::posix_time::microsec_clock::local_time();
		diff = end - temp;
		std::cout << "iteration " << iter << " done in " << diff.total_milliseconds() << " seconds" << std::endl;
		temp = end;

		if (should_print_bounds) {
			bounds_log << upper_bound + tolerance << ";" << upper_bound << ";"
			<< upper_bound - tolerance << ";" << lower_bound << std::endl;
		}
	}

	//RunLastIterationWithKnownInflow();

	//StopSimulation();
	diff = end - begin;
	std::cout << "finished in " << iter << " iterations in " << diff.total_milliseconds() << " seconds" << std::endl;
	total_time_spent += diff.total_milliseconds();

	//results_log.open("../../output/results.txt", std::ofstream::app);
	//results_log << iter << ";";
	//results_log.close();

	if (should_print || should_print_send_log || should_print_time_data) {
		problem_log.open("../../output/problem.txt", std::fstream::app);
		problem_log << iter << " " << size << std::endl;
		problem_log.close();
	}
	StoreResults(iter, diff.total_milliseconds());
}

void Master::ForwardSolve() {
	State* current_state;
	int index;
	int flag;
	int type;
	int num_states = (num_stages - 1)*num_branches;
	int states_sent = 0;
	int states_calculated = 0;
	int available_slaves;
	State* solved_state = NULL;
	MPI_Status status;
	std::queue<State*> state_queue;
	boost::posix_time::ptime tick;
	boost::posix_time::time_duration diff;
	int* lol = new int[num_slaves];
	std::fill_n(lol, num_slaves, 0);

	//slave_queue.clear();

	ForwardSolveRoot();

//	for (int i = 1; i < num_slaves + 1; i++) {
//		forward_slave_queue.push_back(i);
//	}
//	for (int i = 1; i < num_slaves + 1; i++) {
//		forward_slave_queue.push_back(i);
//	}

//	for (int i = 0; i < num_branches; i++) {
//		my_queue.push(root->branches[i]);
//	}
	
	//index = 1;
	if (slave_queue.size() > num_slaves) {
		available_slaves = num_slaves;
	}
	else {
		available_slaves = slave_queue.size();
	}

	for (int i = 0; i < available_slaves; i++) {
		index = slave_queue.front();
		slave_queue.pop_front();
		SendMessage(1, i, index, FORWARD);
	}

	if (2*available_slaves > num_branches) {
		for (int i = available_slaves; i < num_branches; i++) {
			index = slave_queue.front();
			slave_queue.pop_front();
//			MPI_Wait(&array_of_requests[index - 1], &status);
			SendMessage(1, i, index, FORWARD);
//			index++;
			//if (index == size) {
			//	index = 1;
			//}
		}
		states_sent= num_branches;
	}
	else {
		for (int i = available_slaves; i < 2*available_slaves; i++) {
			index = slave_queue.front();
			slave_queue.pop_front();
			SendMessage(1, i, index, FORWARD);
//			index++;
			//if (index == size) {
			//	index = 1;
			//}
		}
		states_sent = 2*available_slaves;

		for (int i = 2*available_slaves; i < num_branches; i++) {
			state_queue.push(root->branches[i]);
		}
	}

	//for (int i = 0; i < num_branches; i++) {
	//	SendMessage(1, i, index, FORWARD);
	//	index++;
	//	if (index == size) {
	//		index = 1;
	//	}
	//}

	while (states_calculated < num_states) {
		if (!state_queue.empty() && !slave_queue.empty()){
//			Testany(index, flag, MPI_STATUS_IGNORE);
//			std::cout << "test: " << flag << std::endl;
			index = slave_queue.front();
//			std::cout << index << std::endl;
			slave_queue.pop_front();
			lol[index-1]++;
			current_state = state_queue.front();
			//				UpdateStateInflow(current_state);
			state_queue.pop();
			//				std::cout << current_state->branch << " : " << index << std::endl;
			tick = boost::posix_time::microsec_clock::local_time();

			SendMessage(current_state->stage, current_state->branch, index, FORWARD);

			diff = boost::posix_time::microsec_clock::local_time() - tick;

			if (should_print_send_log) {
				send_time_log << diff.total_milliseconds() << std::endl;
			}


			time_spent_sending_forward_messages += diff.total_milliseconds();

			states_sent++;
		}
		else{
			MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);

			tick = boost::posix_time::microsec_clock::local_time();

			ReceiveAnyMessage(status, type, solved_state);

			diff = boost::posix_time::microsec_clock::local_time() - tick;

			if (should_print_send_log) {
				if (type == FORWARD) {
					receive_time_log << diff.total_milliseconds() << std::endl;
				}
			}


			time_spent_receiving_forward_messages += diff.total_milliseconds();

			if (type == FORWARD) {
				if ((solved_state->next_state) != NULL) { // if not last stage
					state_queue.push(solved_state->next_state);
					solved_state = NULL;
				}
				slave_queue.push_back(status.MPI_SOURCE);
				states_calculated++;
			}
			else {
				slave_queue.push_back(status.MPI_SOURCE);
				std::cout << "does this happen?: " << status.MPI_SOURCE << std::endl;
			}
		}
	}
	for (int i = 0; i < num_slaves; i++) {
		std::cout << lol[i] << " ; ";
	}
	std::cout << std::endl;
}

void Master::ForwardSolveRoot() {
	int index;
	int type;
	int it = 0;
	bool received_forward_root = false;
	MPI_Status status;
	State* temp;
//	Waitany(index, MPI_STATUS_IGNORE);
	index = slave_queue[0];
	SendMessage(0, 0, index, FORWARD);
	while (!received_forward_root) {
		MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
		if (status.MPI_TAG == 1000000000 + iter*1000000) {
			ReceiveForwardMessage(status, temp);
			received_forward_root = true;
		}
		else {
			ReceiveAnyMessage(status, type, temp);
			slave_queue.push_back(status.MPI_SOURCE);
		}
	}
	//MPI_Probe(index, 1000000000 + iter*1000000, MPI_COMM_WORLD, &status);
	
	for (int i = 0; i < num_branches; i++){
		root->branches[i]->initial_reservoir = root->reservoir_solution;
	}
}

void Master::UpdateStateInflow(State* current_state){
//	current_state->relative_inflow = inflow_data->create_inflow(current_state->stage, current_state->previous_state->relative_inflow);
}

void Master::ReceiveForwardMessages(int stage) {
	State* state;
	int it;
	int num_elements;
	for (int i = 0; i < num_branches; i++) {
		it = 0;
		state = stage_data[stage]->getState(i);
		num_elements = num_modules + 2;
		incoming_message = new double[num_elements];
		MPI_Recv(incoming_message, num_elements, MPI_DOUBLE, MPI_ANY_SOURCE, state->tag + iter*1000000, MPI_COMM_WORLD, MPI_STATUS_IGNORE);

		if (should_print) {
			for (int i = 0; i < num_elements; i++) {
				receive_forward_log << incoming_message[i] << " ";
			}
			receive_forward_log << std::endl;
		}

		for (int i = 0; i < num_modules; i++) {
			state->solution[i] = incoming_message[it];
			it++;
		}
		state->objective_value = incoming_message[it]; it++;
		state->future_cost = incoming_message[it]; it++;
		state->current_cost = state->objective_value - state->future_cost;

		GetReservoirSolution(state);
		GetOutflow(state);
		GetSpillage(state);

		if (stage < (num_stages - 1)) {
			state->next_state->initial_reservoir = state->reservoir_solution;
		}

		delete[] incoming_message;

	}
}

void Master::Testany(int &index, int &flag, MPI_Status* status){
	for (int i = 0; i < num_slaves; i++) {
		if (array_of_requests[i] == MPI_REQUEST_NULL) {
			flag = 1;
			index = i + 1;
			std::cout << index << std::endl;
			return;
		}
	}
	MPI_Testany(num_slaves, array_of_requests, &index, &flag, status);
	index++;
}

void Master::Waitany(int &index, MPI_Status* status){
	for (int i = 0; i < num_slaves; i++) {
		if (array_of_requests[i] == MPI_REQUEST_NULL) {
			index = i + 1;
			return;
		}
	}
	MPI_Waitany(num_slaves, array_of_requests, &index, status);
	index++;
}

void Master::GetReservoirSolution(State* state) {
	int variable = num_thermal + 1; // skip future cost and thermal
	for (int i = 0; i < num_modules; i++) {
		state->reservoir_solution[i] = state->solution[variable];
		variable += 3; // skip outflow and spillage
	} 
}

void Master::GetOutflow(State* state) {
	int variable = num_thermal + 2; // skip future cost, thermal and first reservoir storage
	for (int i = 0; i < num_modules; i++) {
		state->outflow[i] = state->solution[variable];
		variable += 3; // skip spillage and storage
	} 
}

void Master::GetSpillage(State* state) {
	int variable = num_thermal + 3; // skip future cost, thermal, first storage and first outflow
	for (int i = 0; i < num_modules; i++) {
		state->spillage[i] = state->solution[variable];
		variable += 3; // skip storage and outflow
	} 
}

void Master::BackwardSolve() {
	int index;
	int branches_sent = 0;
	int flag;
	int available_slaves;
	MPI_Status status;

	boost::posix_time::ptime tick;
	boost::posix_time::time_duration diff;

	if (slave_queue.size() > num_slaves) {
		available_slaves = num_slaves;
	}
	else {
		available_slaves = slave_queue.size();
	}

	for (int i = 0; i < available_slaves; i++) {
		index = slave_queue.front();
		slave_queue.pop_front();
		SendMessage(num_stages - 1, i, index, BACKWARD);
		branches_sent++;
	}

	if (2*available_slaves > num_branches) {
		for (int i = available_slaves; i < num_branches; i++) {
//			MPI_Wait(&array_of_requests[index - 1], &status);
			index = slave_queue.front();
			slave_queue.pop_front();
			SendMessage(num_stages - 1, i, index, BACKWARD);
			branches_sent++;
		}
//		states_sent= num_branches;
	}
	else {
		for (int i = available_slaves; i < 2*available_slaves; i++) {
			index = slave_queue.front();
			slave_queue.pop_front();
//			std::cout << index << std::endl;
			SendMessage(num_stages - 1, i, index, BACKWARD);
			branches_sent++;
		}
//		states_sent = 2*num_slaves;

		for (int i = 2*available_slaves; i < num_branches; i++) {
			MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
			ReceiveBackwardMessage(status);
//			std::cout << status.MPI_SOURCE << std::endl;
//			slave_queue.push_back(status.MPI_SOURCE);
			SendMessage(num_stages - 1, i, status.MPI_SOURCE, BACKWARD);
			branches_sent++;
		}
	}
//	slave_queue.clear();
	ReceiveBackwardMessages(num_stages - 1);


//	index = 1;
//	for (int i = num_slaves; i < num_branches; i++) {
////		MPI_Wait(&array_of_requests[index - 1], &status);
//		SendMessage(num_stages - 1, i, index, BACKWARD);
//		index++;
//		if (index == size) {
//			index = 1;
//		}
//	}

	//for (int i = 0; i < num_branches; i++) {
	//	SendMessage(num_stages - 1, i, index, BACKWARD);
	//	index++;
	//	if (index==size) {
	//		index = 1;
	//	}
	//}

	

	for (int i = num_stages - 2; i > 0; i--) {
		branches_sent = 0;
		while (branches_sent < num_branches) {
//			Testany(index, flag, MPI_STATUS_IGNORE);
			flag = !slave_queue.empty();
			if (flag){
				index = slave_queue.front();
				slave_queue.pop_front();
//				std::cout << index << std::endl;
				tick = boost::posix_time::microsec_clock::local_time();
				SendMessage(i, branches_sent, index, BACKWARD);
				diff = boost::posix_time::microsec_clock::local_time() - tick;
				time_spent_sending_backward_messages += diff.total_milliseconds();
				branches_sent++;
			}
			MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &status);
			while (flag) {
				ReceiveBackwardMessage(status);
//				std::cout << status.MPI_SOURCE << std::endl;
				slave_queue.push_back(status.MPI_SOURCE);
				MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &status);
			}
		}
		
		ReceiveBackwardMessages(i);

	}
}

void Master::SendMessage(int stage, int branch, int slave, int type) {
	int tag;
	int num_elements = 0;
	State* state = stage_data[stage]->getState(branch);
	if (type == FORWARD) {
		//send cuts and initial reservoir
		num_elements = stage_data[stage]->getNumNewCuts(slave)*(num_modules + 1) + num_modules;
	}
	else {
		//send cuts and initial reservoir
		num_elements = stage_data[stage]->getNumNewCuts(slave)*(num_modules + 1) + num_modules;
	}
	MPI_Wait(&array_of_requests[slave - 1], MPI_STATUS_IGNORE);
	delete[] outgoing_messages[slave - 1];
	outgoing_messages[slave - 1] = new double[num_elements];
	CreateOutgoingMessage(stage_data[stage], state, slave, type);
	double diff = double(clock() - start_time)/CLOCKS_PER_SEC;

	if (should_print_send_log) {
		send_log  << type << " " << iter << " " << stage << " " << branch << " " << stage_data[stage]->getNumCuts() << " "
			 << diff <<  " " << slave << std::endl;
	}

	type==FORWARD? tag = state->tag : tag = state->tag - 1000000000;
	tag += iter*1000000;
	MPI_Isend(outgoing_messages[slave-1], num_elements, MPI_DOUBLE, slave, tag, MPI_COMM_WORLD, &array_of_requests[slave-1]);
//	MPI_Send(outgoing_messages[slave-1], num_elements, MPI_DOUBLE, slave, tag, MPI_COMM_WORLD);
}

void Master::CreateOutgoingMessage(StageData* stage_data, State* state, int slave, int type){
	int it = 0;
	for (int i = 0; i < num_modules; i++){
		outgoing_messages[slave - 1][it] = state->initial_reservoir[i]; it++;
	}
	for (int i = 0; i < stage_data->getNumNewCuts(slave); i++){
		for (int j = 0; j < num_modules; j++){
			outgoing_messages[slave - 1][it] = stage_data->getNewCut(i, slave)->phi[j];
			it++;
		}
		outgoing_messages[slave - 1][it] = stage_data->getNewCut(i, slave)->delta;
		it++;
	}

	stage_data->UpdateNumCuts(slave);
}

void Master::ReceiveBackwardMessages(int stage) {
	int flag = 1;
	MPI_Status status;
	while ( stage_data[stage - 1]->getNumCuts() < (num_branches*iter + min_new_cuts) ) {
		ReceiveBackwardMessage(status);
		slave_queue.push_back(status.MPI_SOURCE);
	}
	while (flag) {
		MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &status);
		if (flag) {
				ReceiveBackwardMessage(status);
				slave_queue.push_back(status.MPI_SOURCE);
		}
	}
	if (stage == 1) {
		while ( stage_data[stage - 1]->getNumCuts() < (num_branches*(iter + 1)) ) {
			ReceiveBackwardMessage(status);
			slave_queue.push_back(status.MPI_SOURCE);
		}
	}
}

void Master::ReceiveForwardMessage(MPI_Status &status, State* &solved_state) {
	int num_elements;
	MPI_Get_count(&status, MPI_DOUBLE, &num_elements);
	MPI_Recv(incoming_message, num_elements, MPI_DOUBLE, status.MPI_SOURCE, status.MPI_TAG, MPI_COMM_WORLD, &status);
	
	int received_iteration = (status.MPI_TAG - 1000000000)/1000000;
	int branch = (status.MPI_TAG - 1000000000 - 1000000*received_iteration)/1000;
	int stage = status.MPI_TAG - 1000000000 - 1000000*received_iteration - 1000*branch;
	solved_state = stage_data[stage]->getState(branch);

	if (should_print) {
		receive_forward_log << iter << " " << stage << " " << branch << " " << status.MPI_SOURCE << " ";
		for(int i = 0; i < num_elements; i++) {
			receive_forward_log << incoming_message[i] << " ";
		}
		receive_forward_log << std::endl;
	}

	AddSolution(solved_state);
}

void Master::ReceiveBackwardMessage(MPI_Status &status) {
	boost::posix_time::ptime tick;
	boost::posix_time::time_duration diff;
	int num_elements = num_modules + 1;

	tick = boost::posix_time::microsec_clock::local_time();

	MPI_Recv(serialized_cut, num_elements, MPI_DOUBLE, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
	diff = boost::posix_time::microsec_clock::local_time() - tick;
	time_spent_receiving_backward_messages += diff.total_milliseconds();

	AddCut(status.MPI_TAG);
}

void Master::ReceiveAnyMessage(MPI_Status &status, int &type, State* &solved_state) {
	if (status.MPI_TAG > 1000) {
		type = FORWARD;
		ReceiveForwardMessage(status, solved_state);
	}
	else {
		type = BACKWARD;
		ReceiveBackwardMessage(status);
	}
}

void Master::AddSolution(State* state){
	int it = 0;
	for (int i = 0; i < num_modules; i++) {
		state->reservoir_solution[i] = incoming_message[it];
		it++;
	}
	state->objective_value = incoming_message[it]; it++;
	state->future_cost = incoming_message[it]; it++;
	state->current_cost = state->objective_value - state->future_cost;

	if ( (state->stage < (num_stages - 1)) && (state->stage != 0) ) {
		state->next_state->initial_reservoir = state->reservoir_solution;
	}
}

void Master::AddCut(int stage) {
	Cut* new_cut = new Cut(num_modules);
	new_cut->Initialize();
	for (int i = 0; i < num_modules; i++) {
		new_cut->phi[i] = serialized_cut[i];
	}
	
	new_cut->delta = *(serialized_cut + num_modules);
	stage_data[stage-1]->addCut(new_cut);
}

void Master::FindBounds() {
	upper_bound = 0.0;
	upper_bound_var = 0.0;
	double upper_bound_temp;
	State* current_state;
	for (int i = 0; i < num_branches; i++){
		current_state = root->branches[i];
		for (int j = 1; j < num_stages - 1; j++){
			upper_bound += current_state->current_cost;
			current_state = current_state->next_state;
		}
		upper_bound += current_state->objective_value; // last stage add "future cost"
	}
	upper_bound = upper_bound/num_branches;
	upper_bound += root->current_cost;

	for (int i = 0; i < num_branches; i++){
		current_state = root->branches[i];
		upper_bound_temp = root->current_cost;
		for (int j = 1; j < num_stages - 1; j++){
			upper_bound_temp += current_state->current_cost;
			current_state = current_state->next_state;
		}
		upper_bound_temp += current_state->objective_value; // last stage add "future cost"

		upper_bound_var += (upper_bound_temp - upper_bound)*(upper_bound_temp - upper_bound);
	}
	if (num_branches > 1) {
		upper_bound_var = upper_bound_var/(num_branches - 1);
	}
	lower_bound = root->objective_value;
}

void Master::StopSimulation(){
	int stop = 911;
	for (int i = 1; i < size; i++) {
		MPI_Send(&stop, 1, MPI_INT, i, 911, MPI_COMM_WORLD);
	}
}

void Master::StoreResults(int in_num_iterations, double in_time) {
	results.num_iterations.push_back(in_num_iterations);
	results.time.push_back(in_time);
}

void Master::RunLastIterationWithKnownInflow() {

	boost::posix_time::ptime begin, temp, tick, end;
	boost::posix_time::time_duration diff;

	iter--;

	begin = boost::posix_time::microsec_clock::local_time();
	temp = begin;
	tick = boost::posix_time::microsec_clock::local_time();

	ForwardSolve();

	diff = boost::posix_time::microsec_clock::local_time() - tick;
	time_spent_in_forward_cycle += diff.total_milliseconds();
	tick = boost::posix_time::microsec_clock::local_time();

	BackwardSolve();

	diff = boost::posix_time::microsec_clock::local_time() - tick;
	time_spent_in_backward_cycle += diff.total_milliseconds();

	if (should_print) {
		for (int i = 0; i < num_stages - 1; i++) {
			main_log << stage_data[i]->getNumCuts() << ";";
		}
		main_log << std::endl;
	}


	FindBounds();
	tolerance = 1.645*sqrt(upper_bound_var/num_branches);
	error = upper_bound - lower_bound;
	end = boost::posix_time::microsec_clock::local_time();
	diff = end - temp;
	iter += 2;
	std::cout << "iteration " << iter << " done in " << diff.total_milliseconds() << " seconds" << std::endl;
	temp = end;

	if (should_print) {
		bounds_log << upper_bound + tolerance << ";" << upper_bound << ";"
			<< upper_bound - tolerance << ";" << lower_bound << std::endl;
	}

}

void Master::Reset() {
	ResetStageData();
	CreateInflowSamples();
	CreateForwardInflow();
	FlushMessages();
	SendInflowSamples();
	SendForwardInflow();

	slave_queue.clear();
	for (int i = 0; i < num_slaves; i++) {
		slave_queue.push_back(i+1);
	}
	for (int i = 0; i < num_slaves; i++) {
		slave_queue.push_back(i+1);
	}
}

void Master::FlushMessages() {
	double dummy = 1.0;
	int flag = 1;
	int messages_left = 2*num_slaves - slave_queue.size();
	MPI_Status status;

	while (messages_left) {
		std::cout << messages_left << std::endl;
		MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
		ThrowAwayReceivedMessage(status);
		messages_left--;
	}

	/*

	while (flag) {
		MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &status);
		if (flag) {
				ThrowAwayReceivedMessage(status);
		}
	}
	*/
	
	for (int i = 1; i < size; i++) {
		MPI_Send(&dummy, 1, MPI_DOUBLE, i, 666, MPI_COMM_WORLD);
	}
	
	//TODO FIKS SLIK AT MAN ER GARANTERT AT DETTE SKJER
}

void Master::ThrowAwayReceivedMessage(MPI_Status status) {
	int num_elements;
	MPI_Get_count(&status, MPI_DOUBLE, &num_elements);
	double* dummy = new double[num_elements];
	MPI_Recv(dummy, num_elements, MPI_DOUBLE, status.MPI_SOURCE, status.MPI_TAG, MPI_COMM_WORLD, &status);
	delete[] dummy;
}

void Master::ResetStageData() {
	for (int i = 0; i < num_stages; i++) {
		stage_data[i]->Reset();
	}
}

void Master::Finalize(){
	if (should_print_bounds) {
		bounds_log.close();
	}

	if (should_print) {
		main_log.close();
		receive_forward_log.close();
		receive_backward_log.close();
		cuts_log.close();
		problem_log.close();
	}

	if (should_print_send_log) {
		send_log.close();
		send_time_log.close();
		receive_time_log.close();
		problem_log.close();
	}
}

void Master::PrintResults(double average_iter, double average_time) {
	std::stringstream temp_string;
	std::string s;
	const char* file;
	std::stringstream temp_string2;
	std::string s2;
	const char* file2;
	std::stringstream temp_string3;
	std::string s3;
	const char* file3;
	std::stringstream temp_string4;
	std::string s4;
	const char* file4;

	if (should_print || should_print_send_log || should_print_time_data) {
		std::ofstream label;
		temp_string << "../../output/label.txt";
		s = temp_string.str();
		file = s.c_str();
		label.open(file);
		label << num_stages << "-" << num_substages << "-" << num_branches << "-" << num_samples;
		label << "-" << min_new_cuts << "-" << should_print << "-" << use_relaxation << "-" << size;
		//	label << ".txt" << std::endl;
		label << std::endl;
		label.close();
	}
	if (!should_print && !should_print_send_log){
		std::ofstream results_log;
		temp_string2 << "../../output/results-";
		temp_string2 << num_stages << "-" << num_substages << "-" << num_branches << "-" << num_samples;
		temp_string2 << "-" << min_new_cuts << "-" << should_print << "-" << use_relaxation << "-" << size;
		temp_string2 << ".txt";
		s2 = temp_string2.str();
		file2 = s2.c_str();
		results_log.open(file2, std::fstream::app);
		for (int i = 0; i < results.num_iterations.size(); i++){
			results_log << results.num_iterations[i] << ";" << results.time[i] << std::endl;
		}
		results_log.close();
	}

	if (should_print_time_data) {
		std::ofstream time_log;
		temp_string3 << "../../output/time_log-";
		temp_string3 << num_stages << "-" << num_substages << "-" << num_branches << "-" << num_samples;
		temp_string3 << "-" << min_new_cuts << "-" << should_print << "-" << use_relaxation << "-" << size;
		temp_string3 << ".txt";
		s3 = temp_string3.str();
		file3 = s3.c_str();
		time_log.open(file3);
		time_log << time_spent_in_forward_cycle << ";" << time_spent_in_backward_cycle << ";" << time_spent_sending_backward_messages << ";" 
			<< time_spent_receiving_backward_messages << ";" << time_spent_receiving_forward_messages << ";"
			<< time_spent_sending_forward_messages << ";" << total_time_spent << std::endl;
		time_log.close();
	}

	if (!should_print && !should_print_send_log){
		std::ofstream results2_log;
		temp_string4 << "../../output/results-";
		temp_string4 << num_stages << "-" << num_substages << "-" << num_branches << "-" << num_samples;
		temp_string4 << "-" << min_new_cuts;
		temp_string4 << ".txt";
		s4 = temp_string4.str();
		file4 = s4.c_str();
		results2_log.open(file4, std::fstream::app);
		for (int i = 0; i < results.num_iterations.size(); i++){
			results2_log << size << ";" << results.num_iterations[i] << ";" << results.time[i] << std::endl;
		}
		results2_log.close();

	}
}