#ifndef MASTER_H
#define MASTER_H
#include <iostream>
#include <fstream>
#include <deque>
#include "powerplant.h"
#include "powersystem.h"
#include "hydromodule.h"
#include "thermalplant.h"
#include "inflow.h"
#include "stagedata.h"
#include "state.h"
#include "mpi.h"
#include <time.h>
#include "substagedata.h"
#include "input.h"
#include "results.h"

class Master{
private:
	int rank;
	const int size;
	int max_num_iterations;

	int num_stages;
	int num_substages;
	int num_modules; 
	int num_thermal;
	int num_branches;
	int num_samples;
	int num_slaves;
	int num_states;
	int num_simulations;
	int min_new_cuts; //minimum amount of cuts to wait for at each stage
	int initial_demand; //input used to generate demand-data
	bool should_print;
	bool should_print_send_log;
	bool should_print_time_data;
	bool should_print_bounds;
	bool use_relaxation;

	PowerSystem* system;
	std::vector<HydroModule*> hydro_modules; 
	std::vector<ThermalPlant*> thermal_plants;
	Demand* demand;

	double lower_bound;
	double upper_bound;
	double upper_bound_var;

	Inflow* inflow_data;
	StageData** stage_data; // stage-dependent data, i.e. cuts.
	Substagedata* substage_data;
	State* root;
	//OsiClpSolverInterface* lp_model;
	//std::default_random_engine generator;
	MPI_Request* array_of_requests;
	double** outgoing_messages;
	double* incoming_message;
	double* serialized_cut;

	clock_t start_time;
	int time_spent_in_forward_cycle;
	int time_spent_in_backward_cycle;
	int time_spent_sending_backward_messages;
	int time_spent_receiving_backward_messages;
	int time_spent_sending_forward_messages;
	int time_spent_receiving_forward_messages;
	int total_time_spent;

	int* slave_processing_amount;
	std::deque<int> slave_queue;

	Results results;

	char* log_identifier;
	std::ofstream main_log;
	std::ofstream problem_log;
	std::ofstream bounds_log;
	//iter stage branch previous_inflow current_inflow
	std::ofstream send_log;
	std::ofstream send_time_log;
	std::ofstream receive_time_log;
	// type iter stage branch numCuts inflow time slave
	std::ofstream cuts_log;
	// iter stage slave cut
	std::ofstream receive_forward_log;
	// iter stage branch slave message
	// message[0] = future cost
	// message[1:num_thermal] = thermal solutuions]
	// message[num_thermal + 1 + 3*k] = reservoir solutions
	// message[num_thermal + 2 + 3*k] = outflow solutions
	// message[num_thermal + 3 + 3*k] = spillage solutuions
	// message[end] = message[num_variables + 1] = objective value
	// objective value - future cost = current const
	std::ofstream receive_backward_log;

	int iter;
	int max_iter;
	double error;
	double tolerance;

	void CreateProblemLog();
	//stages, modules, thermal, branches, samples, slaves, states, min new cuts
	void ReadData(Input*);
	void InitializeStageData();
	void BuildTree();
	void InitializeRoot();
	void InitializeState(State* state);
	void SendMessage(int stage, int branch, int slave, int type);
	void CreateInflowSamples();
	void CreateForwardInflow();
	void SendInflowSamples();
	void SendForwardInflow();
	void CreateOutgoingMessage(StageData*, State* state, int slave, int type);
	void FindBounds();
	void ForwardSolve();
	void ForwardSolveRoot();
	void UpdateStateInflow(State* current_state);
	void ReceiveForwardMessages(int stage);
	void Testany(int &index, int &flag, MPI_Status* status);
	void Waitany(int &index, MPI_Status* status);
	void GetReservoirSolution(State* state);
	void GetOutflow(State* state);
	void GetSpillage(State* state);
	void BackwardSolve();
	void ReceiveBackwardMessages(int state);
	void ReceiveBackwardMessage(MPI_Status &);
	void ReceiveForwardMessage(MPI_Status &, State* &);
	void ReceiveAnyMessage(MPI_Status &, int &type, State* &solved_state);
	void AddSolution(State* state);
	void AddCut(int stage);
	void StopSimulation();
	void StoreResults(int, double);
	void Reset();
	void FlushMessages();
	void ThrowAwayReceivedMessage(MPI_Status);
	void ResetStageData();
	void PrintResults(double, double);
	void RunLastIterationWithKnownInflow();

public:
	Master(PowerSystem* in_system, int rank, int size);

	void Initialize(Input*, Inflow*);
	void RunSimulations();
	void RunSimulation();
	void Finalize();
};

#endif