#include "mysimplexsolvercoinclp.h"
#include "OsiClpSolverInterface.hpp"
#include "CoinPackedVector.hpp"
#include "CoinPackedMatrix.hpp"
#include "cut.h"

MySimplexSolverCoinCLP::MySimplexSolverCoinCLP() {
	lp_model = new OsiClpSolverInterface();
	matrix = new CoinPackedMatrix(false, 0, 0);
	lp_model->setLogLevel(0);
}

void MySimplexSolverCoinCLP::Reset(int num_variables) {
	delete matrix;
	matrix = new CoinPackedMatrix(false, 0, 0);
	matrix->setDimensions(0, num_variables);
	lp_model->reset();
	lp_model->setLogLevel(0);
}

bool MySimplexSolverCoinCLP::InitialSolve() {
	ClpSimplex* model = lp_model->getModelPtr();
	try{
		lp_model->initialSolve();
	}
	catch(...){
		std::cout << "exception!" << std::endl;
	}
	
	if ( !isOptimal() ){
		model->initialPrimalSolve();
	}

	if ( !isOptimal() ){
		model->dual();
	}

	if ( !isOptimal() ){
		model->initialDualSolve();
	}

	if ( !isOptimal() ) {
		std::cout << "optimal not proven!" << std::endl;
		if (model->isProvenDualInfeasible())
			std::cout << "proven dual infeasible!" << std::endl;
		if (model->isProvenPrimalInfeasible())
			std::cout << "proven primal infeasible!" << std::endl;
		if (model->isDualObjectiveLimitReached() )
			std::cout << "dual objective limit reached!" << std::endl;
		if (model->isPrimalObjectiveLimitReached())
			std::cout << "primal objective limit reached!" << std::endl;
		if (model->isAbandoned())
			std::cout << "is abandoned!" << std::endl;
		if (model->isIterationLimitReached())
			std::cout << "iteration limit reached!" << std::endl;
		return false;
	}
	return true;
}

bool MySimplexSolverCoinCLP::Resolve() {
	ClpSimplex* model = lp_model->getModelPtr();

	try{
		lp_model->resolve();
	}
	catch(...){
		std::cout << "exception!" << std::endl;
	}

	if ( !isOptimal() ){
		model->initialDualSolve();
	}

	if ( !isOptimal() ){
		model->initialPrimalSolve();
	}

	if ( !isOptimal() ){
		lp_model->initialSolve();
	}

	if ( !isOptimal() ){
		model->dual();
	}

	if ( !isOptimal() ) {
		std::cout << "optimal not proven!" << std::endl;
		if (model->isProvenDualInfeasible())
			std::cout << "proven dual infeasible!" << std::endl;
		if (model->isProvenPrimalInfeasible())
			std::cout << "proven primal infeasible!" << std::endl;
		if (model->isDualObjectiveLimitReached() )
			std::cout << "dual objective limit reached!" << std::endl;
		if (model->isPrimalObjectiveLimitReached())
			std::cout << "primal objective limit reached!" << std::endl;
		if (model->isAbandoned())
			std::cout << "is abandoned!" << std::endl;
		if (model->isIterationLimitReached())
			std::cout << "iteration limit reached!" << std::endl;
		return false;
	}
	return true;
}

void MySimplexSolverCoinCLP::LoadProblem(int num_variables, std::vector< std::map<int, double> >& constraints, double* col_lb, double* col_ub, 
								  double* objective_coefficients, double* row_lb, double* row_ub) {
	ConvertToCoinPackedMatrix(constraints);
	lp_model->loadProblem(*matrix, col_lb, col_ub, objective_coefficients, row_lb, row_ub);
}

void MySimplexSolverCoinCLP::SetRowBounds(int constraint, double lower_bound, double upper_bound) {
	lp_model->setRowBounds(constraint, lower_bound, upper_bound);
}

void MySimplexSolverCoinCLP::ConvertToCoinPackedMatrix( std::vector< std::map<int, double> >& constraints) {
	CoinPackedVector vector;
	for (int i = 0; i < constraints.size(); i++) {
		vector.clear();
		for(std::map<int,double>::iterator iter = constraints[i].begin(); iter != constraints[i].end(); ++iter) {
			vector.insert(iter->first, iter->second);
		}
		matrix->appendRow(vector);
	}
}

void MySimplexSolverCoinCLP::PrintSolution(const char* file) {
}

bool MySimplexSolverCoinCLP::isOptimal() {
	ClpSimplex* model = lp_model->getModelPtr();
	if ( !(model->dualFeasible()) ){
//		std::cout << "dual infeasible!" << std::endl;
		return false;
	}
	if ( !(model->primalFeasible()) ){
//		std::cout << "primal infeasible!" << std::endl;
		return false;
	}

	if ( !(lp_model->isProvenOptimal()) ){
//		std::cout << "optimal not proven!" << std::endl;
		return false;
	}

	if (lp_model->isProvenDualInfeasible()){
//		std::cout << "proven dual infeasible!" << std::endl;
		return false;
	}
	if (lp_model->isProvenPrimalInfeasible()){
//		std::cout << "proven primal infeasible!" << std::endl;
		return false;
	}
	if (lp_model->isDualObjectiveLimitReached() ){
//		std::cout << "dual objective limit reached!" << std::endl;
		return false;
	}
	if (lp_model->isPrimalObjectiveLimitReached()){
//		std::cout << "primal objective limit reached!" << std::endl;
		return false;
	}
	if (lp_model->isAbandoned()) {
//		std::cout << "is abandoned!" << std::endl;
		return false;
	}
	if (lp_model->isIterationLimitReached()) {
//		std::cout << "iteration limit reached!" << std::endl;
		return false;
	}

	return true;
}

void MySimplexSolverCoinCLP::AddCut(Cut* new_cut, int* reservoir_variables, double lower_bound, int num_substages) {
	double numeric_cut_off = 1e-4;
	int num_modules = new_cut->num_modules;
	int size = num_modules + 1;
	double* phi = new_cut->phi;
	int* columns = new int[size];
	double* elements = new double[size];
	columns[0] = 0;
	elements[0] = 1.0;

	int counter = 0;
	counter++;
	for (int i = 0; i < num_modules; i++) {
		columns[counter] = reservoir_variables[i];
		elements[counter] = -phi[i];
		if (elements[counter] < numeric_cut_off) {
			elements[counter] = 0;
		}
		counter++;
	}

	if (lower_bound < numeric_cut_off) {
		lower_bound = 0;
	}

	lp_model->addRow(size, columns, elements, lower_bound, GetInfinity());
	delete[] columns;
	delete[] elements;
}