#include "slave.h"
#include "hydromodule.h"
#include "reservoir.h"
#include "thermalplant.h"
#include "powerplant.h"
#include "boost/date_time/local_time/local_time.hpp"
#include "boost/thread/thread.hpp"
#include <vector>

Slave::Slave(PowerSystem* in_system, int in_rank, int in_size)
   : system(in_system), rank(in_rank), size(in_size) {
}

void Slave::Initialize(Input* input, Inflow* in_inflow_data) {
	num_prints = 0;
	max_num_prints = 1;

	ReadData(input);
	hydro_modules = system->getHydroModules();
	thermal_plants = system->getThermalPlants();
	num_modules = int(hydro_modules.size());
	num_thermal = int(system->getThermalPlants().size());
	num_samples_backward = num_samples;
	current_iteration = 0;

	if (should_print) {
		InitializeLog();	
	}

	demand = new Demand(initial_demand);
	inflow = new double[num_modules];
	previous_inflow = new double[num_modules];
	
	thermal_solution = new double*[num_substages];
	reservoir_solution = new double*[num_substages];
	outflow = new double*[num_substages];
	spillage = new double*[num_substages];
	tankbil = new double[num_modules];
	for (int i = 0; i < num_substages; i++) {
		thermal_solution[i] = new double[num_thermal];
		reservoir_solution[i] = new double[num_modules];
		outflow[i] = new double[num_modules];
		spillage[i] = new double[num_modules];
	}

	num_variables = 1 + (3*num_modules + num_thermal)*num_substages + num_modules; // future cost, ..., tankbiler
	num_constraints = (1 + num_modules)*num_substages;
	
	substage_data = new Substagedata(num_substages);
	outgoing_message = NULL;
	std::fill_n(inflow, num_modules, 0);
	std::fill_n(previous_inflow, num_modules, 0);
	send_request = MPI_REQUEST_NULL;

	inflow_data = in_inflow_data;
	//inflow_data->SeedRNG(rank*time(NULL));
//	last_sent_time = start_time;
	new_cut = new Cut(num_modules);
	new_cut->Initialize();
	ReceiveInflowSamples();
	ReceiveForwardInflow();
	my_simplex_solver = new MySimplexSolverCoinCLP();

	coefficients = new double[num_variables]; // coefficients in objective function, + 1 for lp_solve purposes
	var_lb = new double[num_variables]; // variables lower bound
	var_ub = new double[num_variables]; // variables upper bound
	con_lb = new double[num_constraints]; // constraints lower bound
	con_ub = new double[num_constraints]; // constraints upper bound
	incoming_message = NULL;
	modified_deltas = NULL;
	cuts = new std::vector<Cut*>[num_stages];

	InitializeVariables();
	InitializeTimeVariables();
	BuildLastStageCut();
	/*
	std::stringstream temp_string;
	temp_string << "../../output/random_stuff";
	temp_string << rank;
	temp_string << ".txt";
	std::string s = temp_string.str();
	const char* file = s.c_str();
	random_stuff_log.open(file);
	*/
}

void Slave::ReadData(Input* input) {
	num_stages = input->num_stages;
	num_branches = input->num_branches;
	num_samples = input->num_samples;
	min_new_cuts = input->min_new_cuts;
	num_substages = input->num_substages;
	initial_demand = input->initial_demand;
	should_print = input->should_print;
	use_relaxation = input->use_relaxation;
	max_num_iterations = input->max_num_iterations;
	should_print_time_data = input->should_print_time_data;
}

void Slave::InitializeLog() {
	std::stringstream temp_string;
	temp_string << "../../output/slave";
	temp_string << rank;
	temp_string << ".txt";
	std::string s = temp_string.str();
	const char* file = s.c_str();
	log.open(file);

	std::stringstream temp_string2;
	temp_string2 << "../../output/tankbil";
	temp_string2 << rank;
	temp_string2 << ".txt";
	std::string s2 = temp_string2.str();
	const char* file2 = s2.c_str();
	tankbil_log.open(file2);

	std::stringstream temp_string3;
	temp_string3 << "../../output/clock_log";
	temp_string3 << rank;
	temp_string3 << ".txt";
	std::string s3 = temp_string3.str();
	const char* file3 = s3.c_str();
	clock_log.open(file3);

	std::stringstream temp_string4;
	temp_string4 << "../../output/dual_variables_log";
	temp_string4 << rank;
	temp_string4 << ".txt";
	std::string s4 = temp_string4.str();
	const char* file4 = s4.c_str();
	dual_variables_log.open(file4);

	std::stringstream temp_string5;
	temp_string5 << "../../output/num_relaxation_cuts_log";
	temp_string5 << rank;
	temp_string5 << ".txt";
	std::string s5 = temp_string5.str();
	const char* file5 = s5.c_str();
	num_relaxation_cuts_log.open(file5);

	std::stringstream temp_string6;
	temp_string6 << "../../output/forward_solve_time_log";
	temp_string6 << rank;
	temp_string6 << ".txt";
	std::string s6 = temp_string6.str();
	const char* file6 = s6.c_str();
	forward_solve_time_log.open(file6);
}

void Slave::InitializeTimeVariables(){
	time_spent_total = 0;
	time_spent_lp_solving = 0;
	time_spent_probing = 0;
	time_spent_receiving_message = 0;
	time_spent_waiting_to_send_message = 0;
	time_spent_sending_message = 0;
	time_spent_forward_lp_solving = 0;
	time_spent_forward_building = 0;
	time_spent_backward_lp_solving = 0;
	time_spent_backward_building = 0;
	time_spent_forward_total = 0;
	time_spent_backward_total = 0;
	time_spent_receiving_messages_2 = 0;
}

void Slave::InitializeVariables() {
	thermal_variables = new int*[num_substages];
	reservoir_variables = new int*[num_substages];
	discharge_variables = new int*[num_substages];
	spillage_variables = new int*[num_substages];
	tankbil_variables = new int[num_modules];

	for (int i = 0; i < num_substages; i++) {
		thermal_variables[i] = new int[num_thermal];
		reservoir_variables[i] = new int[num_modules];
		discharge_variables[i] = new int[num_modules];
		spillage_variables[i] = new int[num_modules];
	}

	int counter = 0;

	future_cost_variable = counter++;

	for (int i = 0; i < num_substages; i++) {
		for (int j = 0; j < num_thermal; j++) {
			thermal_variables[i][j] = counter++;
		}
	}

	for (int i = 0; i < num_substages; i++) {
		for (int j = 0; j < num_modules; j++) {
			reservoir_variables[i][j] = counter++;
			discharge_variables[i][j] = counter++;
			spillage_variables[i][j] = counter++;
		}
	}

	for (int i = 0; i < num_modules; i++) {
		tankbil_variables[i] = counter++;
	}
}

void Slave::RunSimulation() {
	boost::posix_time::ptime tick;
	boost::posix_time::time_duration diff;
	start_time = boost::posix_time::microsec_clock::local_time();

	bool flag = 0;
	ReceiveMessage();
	if (use_relaxation) {
		while (!flag) {
			if (type==FORWARD){
				tick = boost::posix_time::microsec_clock::local_time();
				ForwardSolveRelaxation();
				diff = boost::posix_time::microsec_clock::local_time() - tick;
//				boost::this_thread::sleep( boost::posix_time::milliseconds(30) );
//				std::cout << diff.total_milliseconds() << std::endl;
//				forward_solve_time_log << diff.total_milliseconds() << std::endl;
				time_spent_forward_total += diff.total_milliseconds();
			}
			else if (type==BACKWARD){
				tick = boost::posix_time::microsec_clock::local_time();
				BackwardSolveRelaxation();
				diff = boost::posix_time::microsec_clock::local_time() - tick;
//				std::cout << diff.total_milliseconds() << std::endl;
				time_spent_backward_total += diff.total_milliseconds();
			}
			else if (type==RESET) {
				ReceiveInflowSamples();
				ReceiveForwardInflow();
				ResetCuts();
				current_iteration = 0;
				type = NULL;
			}
			tick = boost::posix_time::microsec_clock::local_time();
			flag = ReceiveMessage();
			diff = boost::posix_time::microsec_clock::local_time() - tick;
			time_spent_receiving_messages_2 += diff.total_milliseconds();
		}
	}
	else{
		while (!flag) {
			if (type==FORWARD){
				ForwardSolve();
			}
			else if (type==BACKWARD){
				BackwardSolve();
			}
			else if (type==RESET) {
				ReceiveInflowSamples();
				ReceiveForwardInflow();
				ResetCuts();
			}
			flag = ReceiveMessage();
		}
	}
}

bool Slave::ReceiveMessage() {
	MPI_Status status;
	int num_elements;
	boost::posix_time::ptime tick;
	boost::posix_time::time_duration diff;

	tick = boost::posix_time::microsec_clock::local_time();
	MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
	diff = boost::posix_time::microsec_clock::local_time() - tick;
	time_spent_probing += diff.total_milliseconds();

	tag = status.MPI_TAG;
	if (tag==911) { // end program
		return 1;
	}
	else if (tag == 666) {
		double dummy;
		type = RESET;
		//MPI_Cancel(&send_request);
		MPI_Wait(&send_request, MPI_STATUS_IGNORE);
		MPI_Recv(&dummy, 1, MPI_DOUBLE, 0, MPI_ANY_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
		return 0;
	}
	else if (tag >= 1000000000) { // forward solve
		type = FORWARD;
	//	boost::this_thread::sleep( boost::posix_time::milliseconds(10) );
		current_iteration = (tag - 1000000000)/1000000;
		branch = (tag - 1000000000 - current_iteration*1000000)/1000;
		stage = tag - 1000000000 - current_iteration*1000000 - branch*1000;
		num_samples = 1;
	}
	else { // backward solve
		type = BACKWARD;
		current_iteration = tag/1000000;
		branch = (tag - current_iteration*1000000)/1000;
		stage = tag - current_iteration*1000000 - branch*1000;
		num_samples = num_samples_backward;
		tag = stage;
	}
	delete[] incoming_message;
	tick = boost::posix_time::microsec_clock::local_time();
	MPI_Get_count(&status, MPI_DOUBLE, &num_elements);
	incoming_message = new double[num_elements];

	MPI_Recv(incoming_message, num_elements, MPI_DOUBLE, 0, status.MPI_TAG, MPI_COMM_WORLD, &status);
	diff = boost::posix_time::microsec_clock::local_time() - tick;
	time_spent_receiving_message += diff.total_milliseconds();

	num_new_cuts = (num_elements - num_modules)/(num_modules + 1);
	current_demand = demand->getDemand(stage);
	ConvertIncomingMessage();
	BuildPreviousInflow();
	return 0;
}

void Slave::ReceiveInflowSamples() {
	double*** new_samples = new double**[max_num_iterations];
	for (int i = 0; i < max_num_iterations; i++) {
		new_samples[i] = new double*[num_stages];
		for (int j = 0; j < num_stages; j++) {
			new_samples[i][j] = new double[num_samples];
		}
	}

	for (int i = 0; i < max_num_iterations; i ++) {
		for (int j = 0; j < num_stages; j++) {
			MPI_Bcast(new_samples[i][j], num_samples_backward, MPI_DOUBLE, 0, MPI_COMM_WORLD);
		}
	}

	inflow_data->ResetNoiseSamples(new_samples);
}

void Slave::ReceiveForwardInflow() {
	double*** new_forward_inflow = new double**[max_num_iterations];
	for (int i = 0; i < max_num_iterations; i ++) {
		new_forward_inflow[i] = new double*[num_stages];
		for (int j = 0; j < num_stages; j++) {
			new_forward_inflow[i][j] = new double[num_branches];
		}
	}

	for (int i = 0; i < max_num_iterations; i++) {
		for (int j = 0; j < num_stages; j++) {
			MPI_Bcast(new_forward_inflow[i][j], num_branches, MPI_DOUBLE, 0, MPI_COMM_WORLD);
		}
		
	}
	inflow_data->ResetForwardInflow(new_forward_inflow);
}

void Slave::ConvertIncomingMessage() {
	Cut* temp_cut;
	double* pos = incoming_message + num_modules;
	initial_reservoir = incoming_message;
	for (int i = 0; i < num_new_cuts; i++) {
		temp_cut= new Cut(num_modules);
		temp_cut->phi = new double[num_modules];
		for (int j = 0; j < num_modules; j++) {
			temp_cut->phi[j] = pos[j];
		}
		pos = pos + num_modules;
		temp_cut->delta = *pos;
		pos++;
		cuts[stage].push_back(temp_cut);
	}
}

void Slave::ForwardSolve(){
	boost::posix_time::ptime tick;
	boost::posix_time::time_duration diff;

	tick = boost::posix_time::microsec_clock::local_time();

	sample = 0;
	BuildForwardInflow();
	BuildModel();
	BuildAllCuts();
	ModifyModel();

	diff = boost::posix_time::microsec_clock::local_time() - tick;
	time_spent_forward_building += diff.total_milliseconds();
	tick = boost::posix_time::microsec_clock::local_time();

	SolveModel();

	diff = boost::posix_time::microsec_clock::local_time() - tick;
	time_spent_forward_lp_solving += diff.total_milliseconds();

	GetSolution();
	SendMessage();
	
	if (should_print) {
		PrintForwardSolution();
	}

	Reset();
}

void Slave::ForwardSolveRelaxation() {
	sample = 0;
	int counter = 0;
	double lower_bound;
	Cut* violated_cut = NULL;
	int violated_cut_number;
	const double* temp_solution;
	boost::posix_time::ptime tick;
	boost::posix_time::time_duration diff;

	BuildForwardInflow();
	BuildModel();
	UpdateReservoirBalance();

	tick = boost::posix_time::microsec_clock::local_time();
	my_simplex_solver->InitialSolve();
	diff = boost::posix_time::microsec_clock::local_time() - tick;
	time_spent_lp_solving += diff.total_milliseconds();
	time_spent_forward_lp_solving += diff.total_milliseconds();

	temp_solution = my_simplex_solver->GetColSolution();
	CalculateModifiedDeltas();

	if (min_new_cuts == num_branches) {
		if (cuts[stage].size()%num_branches!=0 && stage < num_stages - 1) {
			std::cout << "ALERT ALERT, DOES NOT WAIT FOR CORRECT AMOUNT OF CUTS" << std::endl;
			random_stuff_log << cuts[stage].size() << std::endl;
		}
	}

	//violated_cut_number = cuts[stage].size()-1;
	//violated_cut = cuts[stage].back();
	while (ViolatesCutsMostViolatedVersion(temp_solution, violated_cut, violated_cut_number) && counter < cuts[stage].size() ) {
		lower_bound = modified_deltas[violated_cut_number];
		my_simplex_solver->AddCut(violated_cut, reservoir_variables[num_substages-1], lower_bound, num_substages);

		tick = boost::posix_time::microsec_clock::local_time();
		my_simplex_solver->Resolve();
		diff = boost::posix_time::microsec_clock::local_time() - tick;
		time_spent_lp_solving += diff.total_milliseconds();
		time_spent_forward_lp_solving += diff.total_milliseconds();

		temp_solution = my_simplex_solver->GetColSolution();
		counter++;
	}

	if (counter == cuts[stage].size() && counter > 1) {
		std::cout << "violated cuts amount reached" << std::endl;
	}

	
	if (should_print) {
		num_relaxation_cuts_log << counter << std::endl;
	}

	GetSolution();
	SendMessage();
	
	if (should_print) {
		PrintForwardSolution();
	}

	Reset();
}

void Slave::BuildPreviousInflow(){
	for (int i = 0; i < num_modules; i++){
		previous_inflow[i] = inflow_data->get_previous_forward_inflow(current_iteration, stage, branch)*hydro_modules[i]->getReservoir().get_yearly_inflow();
	}
}

void Slave::BuildForwardInflow(){
	for (int i = 0; i < num_modules; i++){
		inflow[i] = inflow_data->get_forward_inflow(current_iteration, stage, branch)*hydro_modules[i]->getReservoir().get_yearly_inflow();
	}
}

void Slave::BackwardSolve(){
	BuildModel();
	BuildAllCuts();
	CreateCut();
	SendMessage();
	Reset();
}

void Slave::BackwardSolveRelaxation(){
	double lower_bound;
	Cut* violated_cut = NULL;
	int violated_cut_number;
	const double* temp_solution;
	int counter = 0;
	std::vector<int> currently_used_cuts;
	boost::posix_time::ptime tick, tick2;
	boost::posix_time::time_duration diff;
	int temp_time_spent_lp_solving = 0;

	if (min_new_cuts == num_branches) {
		if (cuts[stage].size()%num_branches!=0 && stage < num_stages - 1) {
			std::cout << "ALERT ALERT, DOES NOT WAIT FOR CORRECT AMOUNT OF CUTS" << std::endl;
//			random_stuff_log << cuts[stage].size() << std::endl;
		}
	}

	BuildModel();

	tick = boost::posix_time::microsec_clock::local_time();
	for (int i = 0; i < num_samples; i++){
		

		sample = i;
		GetSampleFlow(i);
		CalculateModifiedDeltas();
		UpdateReservoirBalance();
		UpdateCutConstraints(currently_used_cuts);

		

		if (sample == 0) {
			my_simplex_solver->InitialSolve();
		}
		else {
			if (!my_simplex_solver->Resolve()){
				PrintLP();
			}
		}

		temp_solution = my_simplex_solver->GetColSolution();

		counter = 0;
		//violated_cut_number = cuts[stage].size()-1;
		//violated_cut = cuts[stage].back();
		while (ViolatesCutsMostViolatedVersion(temp_solution, violated_cut, violated_cut_number) && counter < cuts[stage].size()) {
			lower_bound = modified_deltas[violated_cut_number];
			my_simplex_solver->AddCut(violated_cut, reservoir_variables[num_substages-1], lower_bound, num_substages);
			my_simplex_solver->Resolve();
			temp_solution = my_simplex_solver->GetColSolution();
			counter++;
			currently_used_cuts.push_back(violated_cut_number);
		}

		if (counter == cuts[stage].size() && counter > 1) {
			std::cout << "violated cuts amount reached" << std::endl;
		}

		

		if (should_print) {
			num_relaxation_cuts_log << i << ";" << counter << ";" << cuts[stage].size() << std::endl;
		}

		GetSolution();

		UpdateCut();

		
	}
	tick2 = boost::posix_time::microsec_clock::local_time();
	diff = tick2 - tick;
	temp_time_spent_lp_solving += diff.total_milliseconds();

	time_spent_lp_solving += temp_time_spent_lp_solving;
	time_spent_backward_lp_solving += temp_time_spent_lp_solving;

	if (should_print) {
		clock_log << stage << ";" << branch << ";" << temp_time_spent_lp_solving << std::endl;
	}

	for (int i = 0; i < num_modules; i++) {
		new_cut->phi[i] = new_cut->phi[i]/num_samples;
	}
	UpdateDelta();
	SendMessage();
	Reset();
}

void Slave::CreateCut() {

	for (int i = 0; i < num_samples; i++){
		sample = i;
		GetSampleFlow(i);
		ModifyModel();
		SolveModel();
		GetSolution();
		UpdateCut();
	}

	for (int i = 0; i < num_modules; i++) {
		new_cut->phi[i] = new_cut->phi[i]/num_samples;
	}
	UpdateDelta();
}

void Slave::BuildModel() {

	std::fill_n(coefficients, num_variables, 0);
	std::fill_n(var_lb, num_variables, 0);
	std::fill_n(var_ub, num_variables, 0);
	std::fill_n(con_lb, num_constraints, 0);
	std::fill_n(con_ub, num_constraints, 0);

	constraints.clear();
	BuildVariables();
	BuildConstraints();

	my_simplex_solver->LoadProblem(num_variables, constraints, var_lb, var_ub, coefficients, con_lb, con_ub);
}

void Slave::BuildVariables() {
	int current_variable = 0;
	double inf = my_simplex_solver->GetInfinity();

	// future cost
	coefficients[current_variable] = 1;
	var_ub[current_variable] = inf;
	current_variable++;

	// thermal costs
	for (int i = 0; i < num_substages; i++) {
		for (int j = 0; j < num_thermal; j++) {
			coefficients[current_variable] = thermal_plants[j]->getCost();
			var_ub[current_variable] = thermal_plants[j]->getMaxCap()/num_substages;
			current_variable++;
		}
	}

	// hydro variables
	for (int i = 0; i < num_substages; i++) {
		for (int j = 0; j < num_modules; j++) {
			var_ub[current_variable++] = hydro_modules[j]->getReservoir().getCapacity(); // storage
			var_ub[current_variable++] = hydro_modules[j]->getPowerPlant().getMaxOutflow()/num_substages; // outflow
			var_ub[current_variable++] = inf;
		}
	}

	// tankbil
	for (int j = 0; j < num_modules; j++) {
		coefficients[current_variable] = 1000;
		var_ub[current_variable++] = inf;
	}
}

void Slave::BuildConstraints() {

	int current_variable;
	int current_constraint;
	int upstream_module_number;
	int upstream_variable;
	int substage_initial_reservoir;
	double rhs = 0.0;
	double efficiency;
	std::vector<HydroModule*> upstream_modules;
	std::map<int, double> temp_constraint;
	

	// power balance
	current_constraint = 0;

	for (int i = 0; i < num_substages; i++) {
		temp_constraint.clear();
		current_variable = i*num_thermal + 1; // skip earlier thermal and future cost
		for (int j = 0 ; j < num_thermal; j++) {
			temp_constraint[current_variable] = 1.00;
			current_variable++;
		}

		current_variable = num_substages*num_thermal + 1 + i*num_modules*3; //skip thermal, future cost and earlier reservoir variables
		for (int j = 0; j < num_modules; j++) {
			current_variable++; //skip storage
			efficiency = hydro_modules[j]->getPowerPlant().getEfficiency();
			temp_constraint[current_variable] = efficiency;
			current_variable += 2; //skip spillage
		}

		con_lb[current_constraint] = substage_data->get_demand(current_demand, i);
		con_ub[current_constraint] = substage_data->get_demand(current_demand, i);
		constraints.push_back(temp_constraint);
		current_constraint++;
	}



	// reservoir balances
	current_variable = num_thermal*num_substages + 1; // skip thermal and future cost
	int tankbil_variabel = 1 + (num_thermal + 3*num_modules)*num_substages; //skip all other variables
	for (int i = 0; i < num_substages; i++) {

		for (int j = 0; j < num_modules; j++) { 
			rhs = 0.0;
			temp_constraint.clear();

			if (i > 0) { //if not first substage
				substage_initial_reservoir = current_variable - 3*num_modules;
				temp_constraint[substage_initial_reservoir] = -1.0;
			}

			temp_constraint[current_variable] = 1.0; current_variable++; // storage
			temp_constraint[current_variable] = 1.0; current_variable++; // outflow
			temp_constraint[current_variable] = 1.0; current_variable++; // spillage
	
			

			upstream_modules = hydro_modules[j]->getUpStreamModules();
			for (size_t us=0; us<upstream_modules.size(); us++) {
				upstream_module_number = upstream_modules[us]->getNr();
				upstream_variable = 1 + num_thermal*num_substages + i*num_modules*3 + 3*(upstream_module_number - 1) + 1; //skip future cost, thermal, reservoir balances from earlier substages
				temp_constraint[upstream_variable] = -1.00; upstream_variable++;
				temp_constraint[upstream_variable] = -1.00; 
			} 
			con_lb[current_constraint] = con_ub[current_constraint] = rhs;

			if (i == 0){
				temp_constraint[tankbil_variabel] = -1.0; tankbil_variabel++; // tankbil
			}
			constraints.push_back(temp_constraint);
			current_constraint++;
		}

	}
}

void Slave::BuildAllCuts() {
	for (int i = 0; i < cuts[stage].size(); i++) {
		my_simplex_solver->AddCut(cuts[stage][i], reservoir_variables[num_substages-1], 0.0, num_substages);
//		my_simplex_solver->AddCuts(cuts[stage], reservoir_variables, 0.0, num_substages);
	}
}

void Slave::BuildLastStageCut() {
	Cut* last_stage_cut = new Cut(num_modules);
	last_stage_cut->Initialize();
	last_stage_cut->phi[0] = -28.0;
	last_stage_cut->phi[1] = -25.5;
	last_stage_cut->phi[2] = -17.5;
	last_stage_cut->phi[3] = -24.0;
	last_stage_cut->phi[4] = -17.5;
	last_stage_cut->phi[5] = -11.0;
	last_stage_cut->phi[6] = -14.0;
	last_stage_cut->phi[7] = -9.0;
	last_stage_cut->phi[8] = -7.0;
	last_stage_cut->phi[9] = -2.0;
	last_stage_cut->phi[10] = -0.5;
	last_stage_cut->phi[11] = -0.0;
	last_stage_cut->delta = 30000;
	cuts[num_stages - 1].push_back(last_stage_cut);
}

void Slave::GetSampleFlow(int sample) {
	relative_inflow = inflow_data->get_inflow_sample(stage, sample, previous_relative_inflow, current_iteration);
	for (int i = 0; i < num_modules; i++){
		inflow[i] = relative_inflow*hydro_modules[i]->getReservoir().get_yearly_inflow();
	}
}

void Slave::ModifyModel() {
	UpdateReservoirBalance();
	UpdateCutConstraints();
}

void Slave::UpdateReservoirBalance() {
	double val;
	int current_constraint = num_substages; //skip power balances
	//first substage
	for (int i = 0; i < num_modules; i++){
		val = substage_data->get_inflow( inflow[i], 0 )  + initial_reservoir[i];
		my_simplex_solver->SetRowBounds(current_constraint, val, val);
		current_constraint++;
	}

	for (int i = 1; i < num_substages; i++) {
		for (int j = 0; j < num_modules; j++){
			val = substage_data->get_inflow( inflow[j], i );
			my_simplex_solver->SetRowBounds(current_constraint, val, val);
			current_constraint++;
		}
	}
}

void Slave::UpdateCutConstraint(int cut) {
	double val;
	double inf = my_simplex_solver->GetInfinity();
	int current_constraint = num_substages*(1 + num_modules) + cut; 

	val = CalculateModifiedDelta(cut);
	my_simplex_solver->SetRowBounds(current_constraint, val, inf);
}

void Slave::UpdateCutConstraints(){
	double val;
	double inf = my_simplex_solver->GetInfinity();
	int current_constraint = num_substages*(1 + num_modules); //skip power balances and reservoir balances
	for (int i = 0; i < cuts[stage].size(); i++){
		val = CalculateModifiedDelta(i);
		my_simplex_solver->SetRowBounds(current_constraint, val, inf); current_constraint++;
	}
}

void Slave::UpdateCutConstraints(std::vector<int>& currently_used_cuts){
	double val;
	double inf = my_simplex_solver->GetInfinity();
	int first_cut = num_substages*(1 + num_modules); //skip power balances and reservoir balances
	int current_cut;

	for (int i = 0; i < currently_used_cuts.size(); i++){
		current_cut = currently_used_cuts.at(i);
		val = CalculateModifiedDelta(current_cut);
		my_simplex_solver->SetRowBounds(first_cut + i, val, inf);
	}
}

double Slave::CalculateModifiedDelta(int cut){
	double delta = cuts[stage][cut]->delta;
	double* phi = cuts[stage][cut]->phi;
	double correlation = inflow_data->get_correlation(stage+1);

	return delta;
}

double Slave::CalculateModifiedDelta(Cut* cut){
	double delta = cut->delta;
	double* phi = cut->phi;
	double correlation = inflow_data->get_correlation(stage + 1);
	double std_dev1 = inflow_data->get_std_dev(stage);
	double std_dev2 = inflow_data->get_std_dev(stage + 1);

	for (int i = 0; i < num_modules; i++){
//		delta += phi[i]*correlation*inflow[i]*std_dev2/std_dev1;
		delta += phi[i]*correlation*(inflow[i] - inflow_data->get_mean(stage)*hydro_modules[i]->getReservoir().get_yearly_inflow());
	}
	return delta;
}

void Slave::CalculateModifiedDeltas() {
	delete[] modified_deltas;
	modified_deltas = new double[cuts[stage].size()];
	for (int i = 0; i < cuts[stage].size(); i++) {
		modified_deltas[i] = CalculateModifiedDelta(i);
	}
}

void Slave::SolveModel() {
	boost::posix_time::ptime tick;
	boost::posix_time::time_duration diff;

	tick = boost::posix_time::microsec_clock::local_time();
	if (sample == 0) {
		my_simplex_solver->InitialSolve();
	}
	else {
		if (!my_simplex_solver->Resolve()){
			PrintLP();
		}
	}

	diff = boost::posix_time::microsec_clock::local_time() - tick;
	time_spent_lp_solving += diff.total_milliseconds();
//	if (should_print) {
//		clock_log << stage << ";" << diff.total_milliseconds() << std::endl;
//	}
}

void Slave::GetSolution() {
	solution = my_simplex_solver->GetColSolution();
	dual_variables = my_simplex_solver->GetRowPrice();
	GetThermalSolution();
	GetReservoirSolution();
	GetDischarge();
	GetSpillage();
	GetTankbilSolution();
	objective_value = my_simplex_solver->GetObjectiveValue();
	future_cost = solution[0];
	current_cost = objective_value - future_cost;
}

void Slave::GetThermalSolution() {
	int variable;
	for (int i = 0; i < num_substages; i++) {
		for (int j = 0; j < num_thermal; j++) {
			variable = thermal_variables[i][j];
			thermal_solution[i][j] = solution[variable];
		}
	}
}

void Slave::GetReservoirSolution() {
	int variable;
	for (int i = 0; i < num_substages; i++) {
		for (int j = 0; j < num_modules; j++) {
			variable = reservoir_variables[i][j];
			reservoir_solution[i][j] = solution[variable];
		}
	}
}

void Slave::GetDischarge() {
	int variable;
	for (int i = 0; i < num_substages; i++) {
		for (int j = 0; j < num_modules; j++) {
			variable = discharge_variables[i][j];
			outflow[i][j] = solution[variable];
		}
	}
}

void Slave::GetSpillage() {
	int variable;
	for (int i = 0; i < num_substages; i++) {
		for (int j = 0; j < num_modules; j++) {
			variable = spillage_variables[i][j];
			spillage[i][j] = solution[variable];
		}
	}
}

bool Slave::GetTankbilSolution(){
	double eps = 10*DBL_MIN;
	int variable;

	for (int i = 0; i < num_modules; i++) {
		variable = tankbil_variables[i];
		tankbil[i] = solution[variable];
	}

	return 0;
}

void Slave::UpdateCut() {
	double correlation = inflow_data->get_correlation(stage+1);
	for (int i = 0; i < num_modules; i++) {
		new_cut->phi[i] += dual_variables[i + num_substages]; //skip power balances
	}
	new_cut->delta += objective_value;
}

void Slave::UpdateDelta(){
	new_cut->delta = new_cut->delta/num_samples;
	double* phi = new_cut->phi;
	double correlation = inflow_data->get_correlation(stage);
	for (int i = 0; i < num_modules; i++) {
		new_cut->delta -= phi[i]*initial_reservoir[i];
	}
}

void Slave::PrintLP() {
	if (num_prints < max_num_prints) {

		std::stringstream temp_string;
		std::stringstream tag_temp;

		temp_string << "../../output/lp_prints/lp";

		if (stage < 100)
			tag_temp << "0";
		if (stage < 10)
			tag_temp << "0";
		tag_temp << stage;
		if (branch < 100)
			tag_temp << "0";
		if (branch < 10)
			tag_temp << "0";
		tag_temp << branch;
		if (sample < 100)
			tag_temp << "0";
		if (sample < 10)
			tag_temp << "0";
		tag_temp << sample;

		temp_string << tag_temp.str();
		std::string s = temp_string.str();
		const char* file = s.c_str();
		my_simplex_solver->WriteLP(file);
		num_prints++;
	}
}

void Slave::PrintSolution() {
		std::stringstream temp_string;
	std::stringstream tag_temp;

	temp_string << "../../output/lp_prints/solution";

	if (stage < 100)
		tag_temp << "0";
	if (stage < 10)
		tag_temp << "0";
	tag_temp << stage;
	if (branch < 100)
		tag_temp << "0";
	if (branch < 10)
		tag_temp << "0";
	tag_temp << branch;
	if (sample < 100)
		tag_temp << "0";
	if (sample < 10)
		tag_temp << "0";
	tag_temp << sample;

	temp_string << tag_temp.str();
	std::string s = temp_string.str();
	const char* file = s.c_str();
	my_simplex_solver->PrintSolution(file);
}

void Slave::PrintForwardSolution(){ //TODO
	log << rank << ";" << stage << ";" << branch << ";";
	log << objective_value << ";" << future_cost << ";" << current_cost << ";";
	for (int i = 0; i < num_substages; i++) {
		for (int j = 0; j < num_thermal; j++) {
			log << thermal_solution[i][j] << ";";
		}
	}
	for (int i = 0; i < num_substages; i++) {
		for (int j = 0; j < num_modules; j++) {
			log << reservoir_solution[i][j] << ";";
			log << outflow[i][j] << ";";
			log << spillage[i][j] << ";";
		}
	}
	log << std::endl;
	dual_variables_log << rank << ";" << stage << ";" << branch << ";";

	int num_duals = num_substages*(num_modules + 1);
	for (int i = 0; i < num_duals; i++) {
		dual_variables_log << dual_variables[i] << ";";
	}

	dual_variables_log << std::endl;

	tankbil_log << rank << ";" << stage << ";" << branch << ";";
	for (int j = 0; j < num_modules; j++) {
		tankbil_log << tankbil[j] << ";";
	}
	tankbil_log << std::endl;
}

void Slave::SendMessage() {
	boost::posix_time::ptime tick, tick2;
	boost::posix_time::time_duration diff;
	tick = boost::posix_time::microsec_clock::local_time();

	MPI_Wait(&send_request, MPI_STATUS_IGNORE);

	tick2 = boost::posix_time::microsec_clock::local_time();
	diff = tick2 - tick;
	time_spent_waiting_to_send_message += diff.total_milliseconds();

	int num_elements = CreateOutgoingMessage();

	tick = boost::posix_time::microsec_clock::local_time();
	MPI_Isend(outgoing_message, num_elements, MPI_DOUBLE, 0, tag, MPI_COMM_WORLD, &send_request);
	tick2 = boost::posix_time::microsec_clock::local_time();
	diff = tick2 - tick;
	time_spent_sending_message += diff.total_milliseconds();

}

int Slave::CreateOutgoingMessage(){
	delete[] outgoing_message;
	int message_size;
	int it = 0;
	if (type == FORWARD) {
		message_size = num_modules + 2; //reservoir solutions, objective value, future cost
		outgoing_message = new double[message_size];
		for (int i = 0; i < num_modules; i++) {
			outgoing_message[it] = reservoir_solution[num_substages - 1][i]; it++;
		}
		outgoing_message[it] = objective_value; it++;
		outgoing_message[it] = future_cost; it++;
		return message_size;
	}
	else if (type == BACKWARD) {
		message_size = num_modules + 1; // one cut
		outgoing_message = new double[message_size];
			for (int i = 0; i < num_modules; i++){
				outgoing_message[it] = new_cut->phi[i];
				it++;
			}
		outgoing_message[it] = new_cut->delta;
		return message_size;
	}
	return 0;
}

void Slave::Reset(){

//	delete[] coefficients;
//	delete[] var_lb;
//	delete[] var_ub;
//	delete[] con_lb;
//	delete[] con_ub;
	new_cut->Reset();
	my_simplex_solver->Reset(num_variables);
}

void Slave::ResetCuts() {
	for (int i = 0; i < num_stages; i++) {
		cuts[i].clear();
	}
}

void Slave::Finalize() {
	if (should_print) {
		log.close();
		tankbil_log.close();
		clock_log.close();
		dual_variables_log.close();
		num_relaxation_cuts_log.close();
		forward_solve_time_log.close();
	}

//	random_stuff_log.close();

	if (should_print_time_data) {
		boost::posix_time::time_duration diff;
		diff = boost::posix_time::microsec_clock::local_time() - start_time;
		time_spent_total = diff.total_milliseconds();

		std::stringstream temp_string;
		std::ofstream time_log;

		temp_string << "../../output/slave_time_log-";
		temp_string << num_stages << "-" << num_substages << "-" << num_branches << "-" << num_samples;
		temp_string << "-" << min_new_cuts << "-" << should_print << "-"<< use_relaxation << "-" << size;
		temp_string << "-" << rank << ".txt";
		std::string s = temp_string.str();
		const char* file = s.c_str();

		time_log.open(file);
		time_log << time_spent_lp_solving << ";" << time_spent_receiving_message << ";" << time_spent_probing << ";" 
				 << time_spent_waiting_to_send_message << ";" << time_spent_sending_message<< ";" 
				 << time_spent_backward_lp_solving<< ";" << time_spent_forward_lp_solving << ";" 
				 << time_spent_forward_total << ";" << time_spent_backward_total << ";" 
				 << time_spent_receiving_messages_2 << ";" << time_spent_total << std::endl;
		time_log.close();
	}
}

bool Slave::ViolatesCuts(const double* temp_solution, Cut*& violated_cut, int& violated_cut_number){
	double tolerance = 1e-6;
	double left_hand_side;
	double right_hand_side;
	for (int i = violated_cut_number - 1; i >= 0; i--) {
		left_hand_side = temp_solution[future_cost_variable];
		right_hand_side = 0.0;
		for (int j = 0; j < num_modules; j++) {
			right_hand_side += cuts[stage][i]->phi[j]*temp_solution[reservoir_variables[num_substages-1][j]];
		}
		right_hand_side += modified_deltas[i];

		if (left_hand_side - right_hand_side < -tolerance) {
			violated_cut = cuts[stage][i];
			violated_cut_number = i;
			return true;
		}
	}
	return false;
}

bool Slave::ViolatesCutsMostViolatedVersion(const double* temp_solution, Cut*& violated_cut, int& violated_cut_number){
	double tolerance = 1e-4;
	double left_hand_side;
	double right_hand_side;
	double max_violation = tolerance;
	bool is_violation = false;
	for (int i = 0; i < cuts[stage].size(); i++) {
		left_hand_side = temp_solution[future_cost_variable];
		right_hand_side = 0.0;
		for (int j = 0; j < num_modules; j++) {
			right_hand_side += cuts[stage][i]->phi[j]*temp_solution[reservoir_variables[num_substages-1][j]];
		}
		right_hand_side += modified_deltas[i];

		if (right_hand_side - left_hand_side > max_violation) {
			max_violation = right_hand_side - left_hand_side;
			violated_cut_number = i;
			violated_cut = cuts[stage][i];
			is_violation = true;
		}
	}
	return is_violation;
}