#ifndef SLAVE_H
#define SLAVE_H

#include <fstream>
#include <iostream>
#include <map>
#include "boost/date_time/local_time/local_time.hpp"
//#include <boost/random.hpp>
//#include <boost/random/normal_distribution.hpp>
#include "powersystem.h"
#include "demand.h"
#include "state.h"
#include "cut.h"
#include "mpi.h"
#include "inflow.h"
#include <time.h>
#include "substagedata.h"
#include "input.h"
#include <float.h>
#include "mysimplexsolvercoinclp.h"
#include <vector>

enum {FORWARD, BACKWARD, RESET};

class Slave
{
private:
	int num_prints;
	int max_num_prints;

	// MPI
	int rank;
	int size;

	// static problem information, initialized at start
	int num_stages;
	int num_substages;
	int num_modules; 
	int num_thermal;
	int num_branches;
	int num_samples;
	int num_samples_backward;
	int num_variables;
	int max_num_iterations;
	int min_new_cuts; //minimum amount of cuts to wait for at each stage
	int initial_demand; //input used to generate demand-data
	bool should_print;
	bool should_print_time_data;
	bool use_relaxation;
	Demand* demand;
	double current_demand;
	PowerSystem* system;
	std::vector<HydroModule*> hydro_modules; 
	std::vector<ThermalPlant*> thermal_plants;
	Inflow* inflow_data;
	Substagedata* substage_data;

	MySimplexSolverCoinCLP* my_simplex_solver;

	//received from master
	int type; // 0 is forward, 1 is backward
	int stage;
	int branch;
	int sample;
	int current_iteration;
	std::vector<Cut*>* cuts;
	double* initial_reservoir;
	double relative_inflow;
	double previous_relative_inflow;

	double* coefficients;
	double* var_lb;
	double* var_ub;
	double* con_lb;
	double* con_ub;
	std::vector< std::map<int, double> > constraints;


	// other
	int tag;
	int num_constraints;
	int num_new_cuts;
	double* inflow;
	double* previous_inflow;
	double* incoming_message;
	double* outgoing_message;

	char* log_identifier;
	std::ofstream log;
	std::ofstream clock_log;
	std::ofstream forward_solve_time_log;
	std::ofstream tankbil_log;
	std::ofstream dual_variables_log;
	std::ofstream num_relaxation_cuts_log;
	std::ofstream random_stuff_log;

	MPI_Request send_request;

	boost::posix_time::ptime start_time;
	int time_spent_total;
	int time_spent_lp_solving;
	int time_spent_probing;
	int time_spent_receiving_message;
	int time_spent_waiting_to_send_message;
	int time_spent_sending_message;
	int time_spent_forward_lp_solving;
	int time_spent_forward_building;
	int time_spent_backward_lp_solving;
	int time_spent_backward_building;
	int time_spent_forward_total;
	int time_spent_backward_total;
	int time_spent_receiving_messages_2;

	// solution
	const double* solution;
	const double* dual_variables;
	double** thermal_solution;
	double** reservoir_solution;
	double** outflow;
	double** spillage;
	double objective_value;
	double future_cost;
	double current_cost;
	double* tankbil;
	Cut* new_cut;

	double* modified_deltas;

	// variables
	int future_cost_variable;
	int** thermal_variables;
	int** reservoir_variables;
	int** discharge_variables;
	int** spillage_variables;
	int* tankbil_variables;

	void ReadData(Input* input);
	void InitializeLog();
	void InitializeTimeVariables();
	void InitializeVariables();
	bool ReceiveMessage();
	void ReceiveInflowSamples();
	void ReceiveForwardInflow();
	void ConvertIncomingMessage();
	void ForwardSolve();
	void ForwardSolveRelaxation();
	void BackwardSolve();
	void BackwardSolveRelaxation();
	void CreateCut();
	void BuildModel();
	void BuildVariables();
	void BuildConstraints();
	void BuildAllCuts();
	void BuildLastStageCut();
	
	void BuildPreviousInflow();
	void BuildForwardInflow();
	void GetSampleFlow(int sample);
	void ModifyModel();
	void UpdateReservoirBalance();
	void UpdateCutConstraint(int cut);
	void UpdateCutConstraints();
	void UpdateCutConstraints(std::vector<int>&);
	double CalculateModifiedDelta(int cut);
	double CalculateModifiedDelta(Cut* cut);
	void CalculateModifiedDeltas();
	void SolveModel();
	void GetSolution();
	void GetThermalSolution();
	void GetReservoirSolution();
	void GetDischarge();
	void GetSpillage();
	bool GetTankbilSolution();
	void UpdateCut();
	void UpdateDelta();
	void PrintLP();
	void PrintSolution();
	void PrintForwardSolution();
	void SendMessage();
	void Reset();
	void ResetCuts();
	int CreateOutgoingMessage();

	bool ViolatesCuts(const double*, Cut*&, int&);
	bool ViolatesCutsMostViolatedVersion(const double*, Cut*&, int&);

public:
	Slave(PowerSystem*, int rank, int size);
	void Initialize(Input*, Inflow*);
	void RunSimulation();
	void Finalize();
};


#endif
