//#include "convertstring.h"
//#include "powerplant.h"
//#include "reservoir.h"
//#include "hydromodule.h"
//#include "thermalplant.h"
//#include "cut.h"
//#include "mpi.h"
//#include <sstream>
//#include <vector>
//#include <map>

template<class MySimplexSolver> 
Slave<MySimplexSolver>::Slave(PowerSystem* in_system, int in_rank, int in_size)
   : system(in_system), rank(in_rank), size(in_size) {
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::Initialize(Input* input, Inflow* in_inflow_data) {
	//temp
	printed = 0;
	counter = 0;


	ReadData(input);
	hydro_modules = system->getHydroModules();
	thermal_plants = system->getThermalPlants();
	num_modules = int(hydro_modules.size());
	num_thermal = int(system->getThermalPlants().size());
	num_samples_backward = num_samples;

	if (should_print) {
		InitializeLog();
	}

	demand = new Demand(initial_demand);
	inflow = new double[num_modules];
	previous_inflow = new double[num_modules];
	
	thermal_solution = new double*[num_substages];
	reservoir_solution = new double*[num_substages];
	outflow = new double*[num_substages];
	spillage = new double*[num_substages];
	tankbil = new double[num_modules];
	for (int i = 0; i < num_substages; i++) {
		thermal_solution[i] = new double[num_thermal];
		reservoir_solution[i] = new double[num_modules];
		outflow[i] = new double[num_modules];
		spillage[i] = new double[num_modules];
	}
	
	substage_data = new Substagedata(num_substages);
	outgoing_message = NULL;
	std::fill_n(inflow, num_modules, 0);
	std::fill_n(previous_inflow, num_modules, 0);
	send_request = MPI_REQUEST_NULL;

	inflow_data = in_inflow_data;
	//inflow_data->SeedRNG(rank*time(NULL));
	start_time = clock();
	last_sent_time = start_time;
	new_cut = new Cut(num_modules);
	new_cut->Initialize();
	ReceiveInflowSamples();
	my_simplex_solver = new MySimplexSolver();
	coefficients = NULL; // coefficients in objective function
	var_lb = NULL; // variables lower bound
	var_ub = NULL; // variables upper bound
	con_lb = NULL; // constraints lower bound
	con_ub = NULL; // constraints upper bound
	incoming_message = NULL;
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::ReadData(Input* input) {
	num_stages = input->num_stages;
	num_branches = input->num_branches;
	num_samples = input->num_samples;
	min_new_cuts = input->min_new_cuts;
	num_substages = input->num_substages;
	initial_demand = input->initial_demand;
	should_print = input->should_print;
	use_tankbil = input->use_tankbil;
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::InitializeLog() {
	std::stringstream temp_string;
	std::stringstream temp_string2;
	std::stringstream temp_string3;
	temp_string << "../../output/slave";
	temp_string2 << "../../output/tankbil";
	temp_string3 << "../../output/clock_log";
	temp_string << rank;
	temp_string2 << rank;
	temp_string3 << rank;
	temp_string << ".txt";
	temp_string2 << ".txt";
	temp_string3 << ".txt";
	std::string s = temp_string.str();
	std::string s2 = temp_string2.str();
	std::string s3 = temp_string3.str();
	const char* file = s.c_str();
	const char* file2 = s2.c_str();
	const char* file3 = s3.c_str();
	log.open(file);
	tankbil_log.open(file2);
	clock_log.open(file3);
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::RunSimulation() {
	bool flag = 0;
	ReceiveMessage();
	while (!flag) {
		if (type==FORWARD){
			ForwardSolve();
		}
		else if (type==BACKWARD){
			BackwardSolve();
		}
		else if (type==RESET) {
			ReceiveInflowSamples();
		}
		flag = ReceiveMessage();
		counter++;
		std::cout << counter << " " << flag << std::endl;
	}
	Finalize();
}

template<class MySimplexSolver>
bool Slave<MySimplexSolver>::ReceiveMessage() {
	MPI_Status status;
	int num_elements;
	double diff;

	MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
	tag = status.MPI_TAG;
	if (tag==911) { // end program
		return 1;
	}
	else if (tag == 666) {
		double dummy;
		type = RESET;
		MPI_Cancel(&send_request);
		MPI_Wait(&send_request, MPI_STATUS_IGNORE);
		MPI_Recv(&dummy, 1, MPI_DOUBLE, 0, MPI_ANY_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
		return 0;
	}
	else if (tag >= 1000000) { // forward solve
		type = FORWARD;
		branch = (tag - 1000000)/1000;
		stage = tag - 1000000 - branch*1000;
		num_samples = 1;
	}
	else { // backward solve
		type = BACKWARD;
		branch = tag/1000;
		stage = tag - branch*1000;
		num_samples = num_samples_backward;
		tag = stage;
	}
	delete[] incoming_message;
	MPI_Get_count(&status, MPI_DOUBLE, &num_elements);
	incoming_message = new double[num_elements];
	diff = double(clock() - last_sent_time)/CLOCKS_PER_SEC;

	if (should_print) {
		clock_log << "ready to receive message after " << diff << " seconds" << std::endl;
	}

	MPI_Recv(incoming_message, num_elements, MPI_DOUBLE, 0, status.MPI_TAG, MPI_COMM_WORLD, &status);
	last_receive_time = clock();
	diff = double(last_receive_time)/CLOCKS_PER_SEC - diff;

	if (should_print) {
		clock_log << "received message after " << diff << " seconds" << std::endl;
	}

	num_cuts = (num_elements - num_modules - 1)/(num_modules + 1);
	if (use_tankbil) {
		num_variables = 1 + (3*num_modules + num_thermal)*num_substages + num_modules; // future cost, ..., tankbiler
	}
	else {
		num_variables = 1 + (3*num_modules + num_thermal)*num_substages;
	}
	num_constraints = (1 + num_modules)*num_substages + num_cuts;
	current_demand = demand->getDemand(stage);
	ConvertIncomingMessage();
	BuildPreviousInflow();
	return 0;
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::ReceiveInflowSamples() {
	double** my_samples = new double*[num_stages];
	for (int i = 0; i < num_stages; i ++) {
		my_samples[i] = new double[num_samples];
	}

	for (int i = 0; i < num_stages; i++) {
		MPI_Bcast(my_samples[i], num_samples, MPI_DOUBLE, 0, MPI_COMM_WORLD);
	}
	inflow_data->ResetNoiseSamples(my_samples);
	delete[] my_samples;
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::ConvertIncomingMessage() {
	cuts = new Cut*[num_cuts];
	double* pos = incoming_message + num_modules;
	initial_reservoir = incoming_message;
	for (int i = 0; i < num_cuts; i++) {
		cuts[i] = new Cut(num_modules);
		cuts[i]->phi = pos;
		pos = pos + num_modules;
		cuts[i]->delta = *pos;
		pos++;
	}
	previous_relative_inflow = *pos;
	if (type == FORWARD) {
		pos++;
		relative_inflow = *pos;
		if (stage == 0){
			relative_inflow = previous_relative_inflow;
			previous_relative_inflow = 0;
		}
	}
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::ForwardSolve(){
	sample = 0;
	BuildForwardInflow();
	BuildModel();
	ModifyModel();
//	PrintLP();
	SolveModel();
//	PrintSolution();
	SendMessage2();
	
	if (should_print) {
		PrintForwardSolution();
	}

	Reset();
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::BuildPreviousInflow(){
	for (int i = 0; i < num_modules; i++){
		previous_inflow[i] = previous_relative_inflow*hydro_modules[i]->getReservoir().get_yearly_inflow();
	}
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::BuildForwardInflow(){
	for (int i = 0; i < num_modules; i++){
		inflow[i] = relative_inflow*hydro_modules[i]->getReservoir().get_yearly_inflow();
	}
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::BackwardSolve(){
	CreateCut();
	SendMessage2();
	Reset();
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::CreateCut() {
	BuildModel();
	for (int i = 0; i < num_samples; i++){
		sample = i;
		GetSampleFlow(i);
		ModifyModel();
		SolveModel();
		UpdateCut();
//		Print();
	}
	for (int i = 0; i < num_modules; i++) {
		new_cut->phi[i] = new_cut->phi[i]/num_samples;
	}
	UpdateDelta();
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::BuildModel() {
	delete[] coefficients;
	delete[] var_lb;
	delete[] var_ub;
	delete[] con_lb;
	delete[] con_ub;

	coefficients = new double[num_variables]; // coefficients in objective function
	var_lb = new double[num_variables]; // variables lower bound
	var_ub = new double[num_variables]; // variables upper bound
	con_lb = new double[num_constraints]; // constraints lower bound
	con_ub = new double[num_constraints]; // constraints upper bound

	std::fill_n(coefficients, num_variables, 0);
	std::fill_n(var_lb, num_variables, 0);
	std::fill_n(var_ub, num_variables, 0);
	std::fill_n(con_lb, num_constraints, 0);
	std::fill_n(con_ub, num_constraints, 0);
	constraints.clear();
	BuildVariables();
	BuildConstraints();
	my_simplex_solver->LoadProblem(num_variables, constraints, var_lb, var_ub, coefficients, con_lb, con_ub);
	//lp_model->assignProblem(matrix, var_lb, var_ub, coefficients, con_lb, con_ub);
	/*lp_model->initialSolve();
	ClpSimplex* model = lp_model->getModelPtr();

	if ( !(lp_model->isProvenOptimal()) ){
		model->primal();
	}

	if ( !(lp_model->isProvenOptimal()) ){
		model->dual();
	}

	if ( !(lp_model->isProvenOptimal()) ) {
		std::cout << "optimal not proven!" << std::endl;
		if (lp_model->isProvenDualInfeasible())
			std::cout << "proven dual infeasible!" << std::endl;
		if (lp_model->isProvenPrimalInfeasible())
			std::cout << "proven primal infeasible!" << std::endl;
		if (lp_model->isDualObjectiveLimitReached() )
			std::cout << "dual objective limit reached!" << std::endl;
		if (lp_model->isPrimalObjectiveLimitReached())
			std::cout << "primal objective limit reached!" << std::endl;
		if (lp_model->isAbandoned())
			std::cout << "is abandoned!" << std::endl;
		if (lp_model->isIterationLimitReached())
			std::cout << "iteration limit reached!" << std::endl;
		Print();
	}
	*/
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::BuildVariables() {
	int current_variable = 0;
	double inf = my_simplex_solver->GetInfinity();

	// future cost
	coefficients[current_variable] = 1;
	var_ub[current_variable] = inf;
	current_variable++;

	// thermal costs
	for (int i = 0; i < num_substages; i++) {
		for (int j = 0; j < num_thermal; j++) {
			coefficients[current_variable] = thermal_plants[j]->getCost();
			var_ub[current_variable] = thermal_plants[j]->getMaxCap();
			current_variable++;
		}
	}

	if (use_tankbil) {
		for (int j = 0; j < num_modules; j++) {
			coefficients[current_variable] = 1000*(num_modules - j);
			var_ub[current_variable] = 10; current_variable++;
		}
	}

	// hydro variables
	for (int i = 0; i < num_substages; i++) {
		for (int j = 0; j < num_modules; j++) {
			var_ub[current_variable] = hydro_modules[j]->getReservoir().getCapacity(); current_variable++; // storage
			var_ub[current_variable] = hydro_modules[j]->getPowerPlant().getMaxOutflow()/num_substages; current_variable++; // outflow
			var_ub[current_variable] = inf; current_variable++;
		}
	}

	// tankbil

}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::BuildConstraints() {

	int current_variable;
	int current_constraint;
	int upstream_module_number;
	int upstream_variable;
	int substage_initial_reservoir;
	double rhs = 0.0;
	double efficiency;
	std::vector<HydroModule*> upstream_modules;
	std::map<int, double> temp_constraint;
	

	// power balance
	current_constraint = 0;

	for (int i = 0; i < num_substages; i++) {
		temp_constraint.clear();
		current_variable = i*num_thermal + 1; // skip earlier thermal and future cost
		for (int j = 0 ; j < num_thermal; j++) {
			temp_constraint[current_variable] = 1.00;
			current_variable++;
		}

		current_variable = num_substages*num_thermal + 1 + num_modules + i*num_modules*3; //skip thermal, future cost, tankbiler and earlier reservoir variables
		for (int j = 0; j < num_modules; j++) {
			current_variable++; //skip storage
			efficiency = hydro_modules[j]->getPowerPlant().getEfficiency();
			temp_constraint[current_variable] = efficiency;
			current_variable += 2; //skip spillage
		}

		con_lb[current_constraint] = con_ub[current_constraint] = substage_data->get_demand(current_demand, i);
		constraints.push_back(temp_constraint);
		current_constraint++;
	}



	// reservoir balances
	current_variable = num_thermal*num_substages + 1 + num_modules; // skip thermal, tankbil and future cost
	int tankbil_variabel = 1 + num_thermal*num_substages;
	for (int i = 0; i < num_substages; i++) {

		for (int j = 0; j < num_modules; j++) { 
			rhs = 0.0;
			temp_constraint.clear();

			if (i > 0) { //if not first substage
				substage_initial_reservoir = current_variable - 3*num_modules;
				temp_constraint[substage_initial_reservoir] = -1.0;
			}

			temp_constraint[current_variable] = 1.0; current_variable++; // storage
			temp_constraint[current_variable] = 1.0; current_variable++; // outflow
			temp_constraint[current_variable] = 1.0; current_variable++; // spillage
	
			

			upstream_modules = hydro_modules[j]->getUpStreamModules();
			for (size_t us=0; us<upstream_modules.size(); us++) {
				upstream_module_number = upstream_modules[us]->getNr();
				upstream_variable = 1 + num_thermal*num_substages + num_modules + i*num_modules*3 + 3*(upstream_module_number - 1) + 1; //skip future cost, thermal, tankbil, reservoir balances from earlier substages
				temp_constraint[upstream_variable] = -1.00; upstream_variable++;
				temp_constraint[upstream_variable] = -1.00; 
			} 
			con_lb[current_constraint] = con_ub[current_constraint] = rhs;

			if (use_tankbil) {
				if (i == 0){
					temp_constraint[tankbil_variabel] = -1.0; tankbil_variabel++; // tankbil
				}
			}
			constraints.push_back(temp_constraint);
			current_constraint++;
		}

	}

	// cuts
	double phi;
	double inf = my_simplex_solver->GetInfinity();
	for (int i = 0; i < num_cuts; i++) {
//		rhs = cuts[i]->delta;
		rhs = 0.0;
		temp_constraint.clear();
		temp_constraint[0] = 1.0; // future cost
		current_variable = 1 + num_thermal*num_substages + (num_substages-1)*num_modules*3 + num_modules; //skip future cost and thermal variables and earlier substages and tankbiler
		for (int j = 0; j < num_modules; j++) {
			phi = cuts[i]->phi[j];
			temp_constraint[current_variable] = -phi;
			current_variable += 3; // skip outflow and spillage
		}
		constraints.push_back(temp_constraint);
		con_lb[current_constraint] = rhs;
		con_ub[current_constraint] = inf;
		current_constraint++;
	}
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::GetSampleFlow(int sample) {
	relative_inflow = inflow_data->get_inflow_sample(stage, sample, previous_relative_inflow);
	for (int i = 0; i < num_modules; i++){
		inflow[i] = relative_inflow*hydro_modules[i]->getReservoir().get_yearly_inflow();
	}
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::ModifyModel() {
	UpdateReservoirBalance();
	UpdateCutConstraints();
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::UpdateReservoirBalance() {
	double val;
	int current_constraint = num_substages; //skip power balances
	//first substage
	for (int i = 0; i < num_modules; i++){
		val = substage_data->get_inflow( inflow[i], 0 )  + initial_reservoir[i];
		my_simplex_solver->SetRowBounds(current_constraint, val, val);
		current_constraint++;
	}

	for (int i = 1; i < num_substages; i++) {
		for (int j = 0; j < num_modules; j++){
			val = substage_data->get_inflow( inflow[j], i );
			my_simplex_solver->SetRowBounds(current_constraint, val, val);
			current_constraint++;
		}
	}
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::UpdateCutConstraints(){
	double val;
	double inf = my_simplex_solver->GetInfinity();
	int current_constraint = num_substages*(1 + num_modules); //skip power balances and reservoir balances
	for (int i = 0; i < num_cuts; i++){
		val = CalculateModifiedDelta(i);
		my_simplex_solver->SetRowBounds(current_constraint, val, inf); current_constraint++;
	}
}

template<class MySimplexSolver>
double Slave<MySimplexSolver>::CalculateModifiedDelta(int cut){
	double delta = cuts[cut]->delta;
	double* phi = cuts[cut]->phi;
	double correlation = inflow_data->get_correlation(stage);
	double std_dev1 = inflow_data->get_std_dev(stage);
	double std_dev2 = inflow_data->get_std_dev(stage + 1);

	for (int i = 0; i < num_modules; i++){
		delta += phi[i]*correlation*inflow[i]*std_dev2/std_dev1;
	}
	return delta;
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::SolveModel() {
	if (sample == 0) {
		my_simplex_solver->InitialSolve();
	}
	else {
		my_simplex_solver->Resolve();
	}
	solution = my_simplex_solver->GetColSolution();
	dual_variables = my_simplex_solver->GetRowPrice();
	if (use_tankbil) {
		tried_to_fix = 0;
		GetTankbilSolution();
	}
	GetThermalSolution();
	GetReservoirSolution();
	GetOutflow();
	GetSpillage();
	objective_value = my_simplex_solver->GetObjectiveValue();
	future_cost = solution[0];
	current_cost = objective_value - future_cost;
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::GetThermalSolution() {
	int variable = 1; // skip future cost
	for (int i = 0; i < num_substages; i++) {
		for (int j = 0; j < num_thermal; j++) {
			thermal_solution[i][j] = solution[variable];
			variable++; 
		}
	}
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::GetReservoirSolution() {
	int variable = num_thermal*num_substages + 1 + num_modules; // skip future cost and thermal and tankbil
	for (int i = 0; i < num_substages; i++) {
		for (int j = 0; j < num_modules; j++) {
			reservoir_solution[i][j] = solution[variable];
			variable += 3; // skip outflow and spillage
		}
	}
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::GetOutflow() {
	int variable = num_thermal*num_substages + 2 + num_modules; // skip future cost, thermal and first reservoir storage and tankbil
	for (int i = 0; i < num_substages; i++) {
		for (int j = 0; j < num_modules; j++) {
			outflow[i][j] = solution[variable];
			variable += 3; // skip outflow and spillage
		}
	}
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::GetSpillage() {
	int variable = num_thermal*num_substages + 3 + num_modules; // skip future cost, thermal, first storage and first outflow and tankbil
	for (int i = 0; i < num_substages; i++) {
		for (int j = 0; j < num_modules; j++) {
			spillage[i][j] = solution[variable];
			variable += 3; // skip outflow and spillage
		}
	}
}

template<class MySimplexSolver>
bool Slave<MySimplexSolver>::GetTankbilSolution(){
	bool tankbil_used = 0;
	double eps = DBL_MIN;
	int current_variable = 1 + num_thermal*num_substages;
	for (int j = 0; j < num_modules; j++) {
		tankbil[j] = solution[current_variable];
		if ( tankbil[j] > eps ) {
			if (should_print){
				tankbil_log << my_simplex_solver->IsProvenOptimal()
							<< ";" << type << ";" << stage << ";" << branch << ";" << j << ";" << current_variable 
							<< ";" << tankbil[j] << ";" << relative_inflow  << ";" << cuts[num_cuts - 1]->phi[j] << std::endl;
			}
			tankbil_used = 1;
		}
		current_variable++;
	}

	if (tankbil_used) {
		std::cout << "tankbil used by processor " << rank << " at stage " << stage << " at branch " << branch << " at sample " << sample;
		if (type == 0){
			std::cout << " in forward cycle." << std::endl;
		}
		else {
			std::cout << " in backward cycle." << std::endl;
		}
		if (!tried_to_fix) {
			my_simplex_solver->InitialSolve();
			tried_to_fix = 1;
			if (!GetTankbilSolution()) {
				std::cout << "fixed." << std::endl;
			}
			else {
				std::cout << "not fixed." << std::endl;
				if (!printed){
					PrintLP();
					printed = 1; //TODO. 
					//GetTankbilSolution();
				}
			}
		}
		
		return 1;
	}
	return 0;
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::UpdateCut() {
	for (int i = 0; i < num_modules; i++) {
		new_cut->phi[i] += dual_variables[i + num_substages]; // skip power balances
	}
	new_cut->delta += objective_value;
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::UpdateDelta(){
	new_cut->delta = new_cut->delta/num_samples;
	double* phi = new_cut->phi;
	double correlation = inflow_data->get_correlation(stage);
	double std_dev1 = inflow_data->get_std_dev(stage - 1);
	double std_dev2 = inflow_data->get_std_dev(stage);

	
	
	
	for (int i = 0; i < num_modules; i++) {
		new_cut->delta -= phi[i]*initial_reservoir[i];
		new_cut->delta -= phi[i]*correlation*previous_inflow[i]*std_dev2/std_dev1;
	}
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::PrintLP() {
	std::stringstream temp_string;
	std::stringstream tag_temp;

	temp_string << "../../output/lp_prints/lp";

	if (stage < 100)
		tag_temp << "0";
	if (stage < 10)
		tag_temp << "0";
	tag_temp << stage;
	if (branch < 100)
		tag_temp << "0";
	if (branch < 10)
		tag_temp << "0";
	tag_temp << branch;
	if (sample < 100)
		tag_temp << "0";
	if (sample < 10)
		tag_temp << "0";
	tag_temp << sample;

	temp_string << tag_temp.str();
	std::string s = temp_string.str();
	const char* file = s.c_str();
	my_simplex_solver->WriteLP(file);
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::PrintSolution() {
		std::stringstream temp_string;
	std::stringstream tag_temp;

	temp_string << "../../output/lp_prints/solution";

	if (stage < 100)
		tag_temp << "0";
	if (stage < 10)
		tag_temp << "0";
	tag_temp << stage;
	if (branch < 100)
		tag_temp << "0";
	if (branch < 10)
		tag_temp << "0";
	tag_temp << branch;
	if (sample < 100)
		tag_temp << "0";
	if (sample < 10)
		tag_temp << "0";
	tag_temp << sample;

	temp_string << tag_temp.str();
	std::string s = temp_string.str();
	const char* file = s.c_str();
	my_simplex_solver->PrintSolution(file);
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::PrintForwardSolution(){
	log << rank << ";" << stage << ";" << branch << ";";
	log << objective_value << ";" << future_cost << ";" << current_cost << ";";
	for (int i = 0; i < num_substages; i++) {
		for (int j = 0; j < num_thermal; j++) {
			log << thermal_solution[i][j] << ";";
		}
	}
	for (int i = 0; i < num_substages; i++) {
		for (int j = 0; j < num_modules; j++) {
			log << reservoir_solution[i][j] << ";";
			log << outflow[i][j] << ";";
			log << spillage[i][j] << ";";
		}
	}
	log << std::endl;
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::SendMessage2() {
	if (stage == 0) {
		// Print();
	}
	// Print();
	double diff = double(clock() - last_receive_time)/CLOCKS_PER_SEC;

	if (should_print) {
		clock_log << "waiting for previous message to have been sent after " << diff << " seconds" << std::endl;
	}


	MPI_Wait(&send_request, MPI_STATUS_IGNORE);
	last_sent_time = clock();
	diff = double( last_sent_time )/CLOCKS_PER_SEC - diff;

	if (should_print) {
		clock_log << "waited for " << diff << " seconds" << std::endl;
	}

	int num_elements = CreateOutgoingMessage();
	MPI_Isend(outgoing_message, num_elements, MPI_DOUBLE, 0, tag, MPI_COMM_WORLD, &send_request);
//	MPI_Send(outgoing_message, num_elements, MPI_DOUBLE, 0, tag, MPI_COMM_WORLD);
//	double diff = double(clock() - start_time)/CLOCKS_PER_SEC;
//	log << rank << " " << stage << " " << branch << " " << diff << std::endl;
}

template<class MySimplexSolver>
int Slave<MySimplexSolver>::CreateOutgoingMessage(){
	delete[] outgoing_message;
	int message_size;
	int it = 0;
	if (type == FORWARD) {
		message_size = num_modules + 2; //reservoir solutions, objective value, future cost
		outgoing_message = new double[message_size];
		for (int i = 0; i < num_modules; i++) {
			outgoing_message[it] = reservoir_solution[num_substages - 1][i]; it++;
		}
		outgoing_message[it] = objective_value; it++;
		outgoing_message[it] = future_cost; it++;
		return message_size;
	}
	else if (type == BACKWARD) {
		message_size = num_modules + 1; // one cut
		outgoing_message = new double[message_size];
			for (int i = 0; i < num_modules; i++){
				outgoing_message[it] = new_cut->phi[i];
				it++;
			}
		outgoing_message[it] = new_cut->delta;
		return message_size;
	}
	return 0;
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::Reset(){
	
	for (int i = 0; i < num_cuts; i++) {
		delete cuts[i];
	}
	
	
	delete[] cuts;
	new_cut->Reset();
	my_simplex_solver->Reset(num_variables);
}

template<class MySimplexSolver>
void Slave<MySimplexSolver>::Finalize() {
	if (should_print) {
		log.close();
		tankbil_log.close();
		clock_log.close();
	}
}